import click

from stoobly_agent.app.settings import Settings

from .decorators.config import ConfigDecorator

settings = Settings.instance()

@click.group(
  epilog="Run 'stoobly-agent feature COMMAND --help' for more information on a command.",
  help="Manage features"
)
@click.pass_context
def feature(ctx):
    pass

@click.group(
  epilog="Run 'stoobly-agent dev_tools COMMAND --help' for more information on a command.",
  help="Toggle whether dev-tools command is enabled",
)
@click.pass_context
def dev_tools(ctx):
  pass

ConfigDecorator(dev_tools, settings, 'cli.features.dev_tools').decorate()

feature.add_command(dev_tools)

@click.group(
  epilog="Run 'stoobly-agent exec COMMAND --help' for more information on a command.",
  help="Toggle whether exec command is enabled",
)
@click.pass_context
def exec(ctx):
  pass

ConfigDecorator(exec, settings, 'cli.features.exec').decorate()

feature.add_command(exec)

@click.group(
  epilog="Run 'stoobly-agent remote COMMAND --help' for more information on a command.",
  help="Toggle whether local or remote storage is used"
)
@click.pass_context
def remote(ctx):
  pass

def __handle_enable_before_action(**kwargs):
  api_key = settings.remote.api_key
  if not api_key or len(api_key) == 0:
    if kwargs.get('api_key'):
      api_key = kwargs['api_key']
    else:
      api_key = click.prompt('Please enter your API key', type=str)

    settings.remote.api_key = api_key

ConfigDecorator(remote, settings, 'cli.features.remote', {
  'enable': {
    'before_action': __handle_enable_before_action,
    'set_options': lambda f: click.option('--api-key')(f), # TODO: support more than one option
  }
}).decorate()

#feature.add_command(remote)


import json

from typing import Union

def decode(data: Union[bytes, str], encoding = None) -> str:
  if isinstance(data, bytes):
    try:
      encoding = encoding if encoding else json.detect_encoding(data)
      return data.decode(encoding)
    except Exception:
      return data.decode('ISO-8859-1')
  else:
    return data


import copy
import pdb

from pathlib import Path
from pprint import pprint
from typing import Dict

import pytest

from stoobly_agent.app.cli.helpers.openapi_endpoint_adapter import (
  OpenApiEndpointAdapter,
)


@pytest.mark.openapi
class TestOpenApiEndpointAdapter():
  @pytest.fixture(scope='class')
  def mock_data_directory_path(self):
    return Path(__file__).parent.parent.parent.parent / 'mock_data'

  @pytest.fixture(scope='class')
  def petstore_file_path(self, mock_data_directory_path):
    path = mock_data_directory_path / "petstore.yaml"
    return path

  @pytest.fixture(scope='class')
  def petstore_expanded_file_path(self, mock_data_directory_path):
    path = mock_data_directory_path / "petstore-expanded.yaml"
    return path

  @pytest.fixture(scope='class')
  def petstore_references_file_path(self, mock_data_directory_path):
    path = mock_data_directory_path / "petstore-references.yaml"
    return path

  @pytest.fixture(scope='class')
  def uspto_file_path(self, mock_data_directory_path):
    path = mock_data_directory_path / "uspto.yaml"
    return path

  @pytest.fixture(scope='class')
  def petstore_swagger_io_file_path(self, mock_data_directory_path):
    path = mock_data_directory_path / "petstore-swagger-io.yaml"
    return path

  @pytest.fixture(scope='class')
  def open_api_endpoint_adapter(self):
    adapter = OpenApiEndpointAdapter()
    return adapter

  class TestWhenAdaptingPestoreExpanded():

    @pytest.fixture(scope='class')
    def expected_v2_get_pets_endpoint(self) -> Dict:
      return {
        'id': 1,
        'method': 'GET',
        'host': 'petstore.swagger.io',
        'port': '443',
        'match_pattern': '/v2/pets',
        'path': '/v2/pets',
        'query_param_names': [
          {
            'endpoint_id': 1,
            'id': 1,
            'inferred_type': 'Array',
            'is_deterministic': True,
            'is_required': False,
            'name': 'tags',
            'query': 'tags',
            'query_param_name_id': None,
          },
          {
            'endpoint_id': 1,
            'id': 2,
            'inferred_type': 'String',
            'is_deterministic': True,
            'is_required': False,
            'name': 'tagsElement',
            'query': 'tags[*]',
            'query_param_name_id': 1,
          },
          {
            'endpoint_id': 1,
            'id': 3,
            'inferred_type': 'Integer',
            'is_deterministic': True,
            'is_required': False,
            'name': 'limit',
            'query': 'limit',
            'query_param_name_id': None,
          }
        ],
        'path_segment_names': [
          {
            'name': 'v2',
            'type': 'static'
          },
          {
            'name': 'pets',
            'type': 'static'
          }
        ],
        'response_param_names': [
          {
            'endpoint_id': 1,
            'id': 1,
            'inferred_type': 'Hash',
            'is_deterministic': True,
            'is_required': False,
            'name': 'Element',
            'query': '[*]',
            'response_param_name_id': None
          },
          {
            'endpoint_id': 1,
            'id': 2,
            'inferred_type': 'String',
            'is_deterministic': True,
            'is_required': True,
            'name': 'name',
            'query': '[*].name',
            'response_param_name_id': 1,
            'values': ['']
          },
          {
            'endpoint_id': 1,
            'id': 3,
            'inferred_type': 'String',
            'is_deterministic': True,
            'is_required': False,
            'name': 'tag',
            'query': '[*].tag',
            'response_param_name_id': 1
          },
          {
            'endpoint_id': 1,
            'id': 4,
            'inferred_type': 'Integer',
            'is_deterministic': True,
            'is_required': True,
            'name': 'id',
            'query': '[*].id',
            'response_param_name_id': 1,
            'values': [0]
          },
        ]
      }

    @pytest.fixture(scope='class')
    def expected_v2_post_pets_endpoint(self) -> Dict:
      return {
        'id': 2,
        'method': 'POST',
        'host': 'petstore.swagger.io',
        'port': '443',
        'match_pattern': '/v2/pets',
        'path': '/v2/pets',
        'body_param_names': [
          {
            'body_param_name_id': None,
            'endpoint_id': 2,
            'id': 1,
            'inferred_type': 'String',
            'is_deterministic': True,
            'is_required': True,
            'name': 'name',
            'query': 'name',
            'values': [''],
          },
          {
            'body_param_name_id': None,
            'endpoint_id': 2,
            'id': 2,
            'inferred_type': 'String',
            'is_deterministic': True,
            'is_required': False,
            'name': 'tag',
            'query': 'tag',
          },
        ],
        'path_segment_names': [
          {
              'name': 'v2',
              'type': 'static',
          },
          {
              'name': 'pets',
              'type': 'static',
          },
        ],
        'response_param_names': [
          {
              'endpoint_id': 2,
              'id': 1,
              'inferred_type': 'String',
              'is_deterministic': True,
              'is_required': True,
              'name': 'name',
              'query': 'name',
              'response_param_name_id': None,
              'values': [''],
          },
          {
              'endpoint_id': 2,
              'id': 2,
              'inferred_type': 'String',
              'is_deterministic': True,
              'is_required': False,
              'name': 'tag',
              'query': 'tag',
              'response_param_name_id': None,
          },
          {
              'endpoint_id': 2,
              'id': 3,
              'inferred_type': 'Integer',
              'is_deterministic': True,
              'is_required': True,
              'name': 'id',
              'query': 'id',
              'response_param_name_id': None,
              'values': [0],
          },
        ],
      }

    @pytest.fixture(scope='class')
    def expected_v2_get_pets_id_endpoint(self) -> Dict:
      return {
        'id': 3,
        'method': 'GET',
        'host': 'petstore.swagger.io',
        'port': '443',
        'path': '/v2/pets/{id}',
        'match_pattern': '/v2/pets/%',
        'aliases': [{'id': 1, 'name': '{id}'}],
        'path_segment_names': [
            {
                'name': 'v2',
                'type': 'static',
            },
            {
                'name': 'pets',
                'type': 'static',
            },
            {
                'name': ':id',
                'type': 'alias',
            },
        ],
        'response_param_names': [
            {
                'endpoint_id': 3,
                'id': 1,
                'inferred_type': 'String',
                'is_deterministic': True,
                'is_required': True,
                'name': 'name',
                'query': 'name',
                'response_param_name_id': None,
                'values': [''],
            },
            {
                'endpoint_id': 3,
                'id': 2,
                'inferred_type': 'String',
                'is_deterministic': True,
                'is_required': False,
                'name': 'tag',
                'query': 'tag',
                'response_param_name_id': None,
            },
            {
                'endpoint_id': 3,
                'id': 3,
                'inferred_type': 'Integer',
                'is_deterministic': True,
                'is_required': True,
                'name': 'id',
                'query': 'id',
                'response_param_name_id': None,
                'values': [0],
            },
        ],
      }

    @pytest.fixture(scope='class')
    def expected_v2_delete_pets_id_endpoint(self) -> Dict:
      return {
        'id': 4,
        'method': 'DELETE',
        'host': 'petstore.swagger.io',
        'path': '/v2/pets/{id}',
        'match_pattern': '/v2/pets/%',
        'port': '443',
        'aliases': [{'id': 1, 'name': '{id}'}],
        'path_segment_names': [
          {
            'name': 'v2',
            'type': 'static',
          },
          {
            'name': 'pets',
            'type': 'static',
          },
          {
            'name': ':id',
            'type': 'alias',
          },
        ],
      }

    def test_adapt_from_file(self, open_api_endpoint_adapter, petstore_expanded_file_path, expected_v2_get_pets_endpoint, expected_v2_post_pets_endpoint, expected_v2_get_pets_id_endpoint, expected_v2_delete_pets_id_endpoint):
      adapter = open_api_endpoint_adapter
      file_path = petstore_expanded_file_path

      endpoints = adapter.adapt_from_file(file_path)

      assert len(endpoints) == 4

      get_pets_endpoint = endpoints[0]
      post_gets_endpoint = endpoints[1]
      get_pets_id_endpoint = endpoints[2]
      delete_pets_id_endpoint = endpoints[3]

      assert get_pets_endpoint == expected_v2_get_pets_endpoint
      assert post_gets_endpoint == expected_v2_post_pets_endpoint
      assert get_pets_id_endpoint == expected_v2_get_pets_id_endpoint
      assert delete_pets_id_endpoint == expected_v2_delete_pets_id_endpoint


  class TestWhenAdaptingUspto():

    @pytest.fixture(scope='class')
    def expected_get_root_https(self) -> Dict:
      return {
        'id': 1,
        'method': 'GET',
        'host': 'developer.uspto.gov',
        'port': '443',
        'path': '/ds-api/',
        'match_pattern': '/ds-api/',
        'path_segment_names': [
          {
            'name': 'ds-api',
            'type': 'static',
          },
        ],
        'response_param_names': [
          {
            'endpoint_id': 1,
            'id': 1,
            'inferred_type': 'Integer',
            'is_deterministic': True,
            'is_required': False,
            'name': 'total',
            'query': 'total',
            'response_param_name_id': None,
          },
          {
            'endpoint_id': 1,
            'id': 2,
            'inferred_type': 'Array',
            'is_deterministic': True,
            'is_required': False,
            'name': 'apis',
            'query': 'apis',
            'response_param_name_id': None,
          },
          {
              'endpoint_id': 1,
              'id': 3,
              'inferred_type': 'Hash',
              'is_deterministic': True,
              'is_required': False,
              'name': 'apisElement',
              'query': 'apis[*]',
              'response_param_name_id': 2,
          },
          {
              'endpoint_id': 1,
              'id': 4,
              'inferred_type': 'String',
              'is_deterministic': True,
              'is_required': False,
              'name': 'apiKey',
              'query': 'apis[*].apiKey',
              'response_param_name_id': 3,
          },
          {
              'endpoint_id': 1,
              'id': 5,
              'inferred_type': 'String',
              'is_deterministic': True,
              'is_required': False,
              'name': 'apiVersionNumber',
              'query': 'apis[*].apiVersionNumber',
              'response_param_name_id': 3,
          },
          {
              'endpoint_id': 1,
              'id': 6,
              'inferred_type': 'String',
              'is_deterministic': True,
              'is_required': False,
              'name': 'apiUrl',
              'query': 'apis[*].apiUrl',
              'response_param_name_id': 3,
          },
          {
              'endpoint_id': 1,
              'id': 7,
              'inferred_type': 'String',
              'is_deterministic': True,
              'is_required': False,
              'name': 'apiDocumentationUrl',
              'query': 'apis[*].apiDocumentationUrl',
              'response_param_name_id': 3,
          },
        ],
      }

    @pytest.fixture(scope='class')
    def expected_get_root_http(self, expected_get_root_https) -> Dict:
      http_endpoint_version = copy.deepcopy(expected_get_root_https)
      http_endpoint_version['id'] = 4
      http_endpoint_version['port'] = '80'
      for response_param_name in http_endpoint_version['response_param_names']:
        response_param_name['endpoint_id'] = 4
      return http_endpoint_version

    @pytest.fixture(scope='class')
    def expected_get_dataset_version_fields_https(self) -> Dict:
      return {
        'id': 2,
        'method': 'GET',
        'host': 'developer.uspto.gov',
        'port': '443',
        'match_pattern': '/ds-api/%/%/fields',
        'path': '/ds-api/{dataset}/{version}/fields',
        'aliases': [{'id': 1, 'name': '{dataset}'}, {'id': 2, 'name': '{version}'}],
        'path_segment_names': [
          {
            'name': 'ds-api',
            'type': 'static',
          },
          {
            'name': ':dataset',
            'type': 'alias',
          },
          {
            'name': ':version',
            'type': 'alias',
          },
          {
            'name': 'fields',
            'type': 'static',
          },
        ],
      }

    @pytest.fixture(scope='class')
    def expected_get_dataset_version_fields_http(self, expected_get_dataset_version_fields_https) -> Dict:
      http_endpoint_version = copy.deepcopy(expected_get_dataset_version_fields_https)
      http_endpoint_version['id'] = 5
      http_endpoint_version['port'] = '80'
      return http_endpoint_version

    @pytest.fixture(scope='class')
    def expected_post_dataset_version_records_https(self) -> Dict:
      return {
        'id': 3,
        'method': 'POST',
        'host': 'developer.uspto.gov',
        'port': '443',
        'path': '/ds-api/{dataset}/{version}/records',
        'match_pattern': '/ds-api/%/%/records',
        'aliases': [{'id': 1, 'name': '{version}'}, {'id': 2, 'name': '{dataset}'}],
        'body_param_names': [
          {
            'body_param_name_id': None,
            'endpoint_id': 3,
            'id': 1,
            'inferred_type': 'String',
            'is_deterministic': True,
            'is_required': True,
            'name': 'criteria',
            'query': 'criteria',
            'values': [''],
          },
          {
            'body_param_name_id': None,
            'endpoint_id': 3,
            'id': 2,
            'inferred_type': 'Integer',
            'is_deterministic': True,
            'is_required': False,
            'name': 'start',
            'query': 'start',
          },
          {
            'body_param_name_id': None,
            'endpoint_id': 3,
            'id': 3,
            'inferred_type': 'Integer',
            'is_deterministic': True,
            'is_required': False,
            'name': 'rows',
            'query': 'rows',
          }
        ],
        'path_segment_names': [
          {
            'name': 'ds-api',
            'type': 'static',
          },
          {
            'name': ':dataset',
            'type': 'alias',
          },
          {
            'name': ':version',
            'type': 'alias',
          },
          {
            'name': 'records',
            'type': 'static',
          },
        ],
     'response_param_names': [
          {
              'endpoint_id': 3,
              'id': 1,
              'inferred_type': 'Hash',
              'is_deterministic': True,
              'is_required': False,
              'name': 'Element',
              'query': '[*]',
              'response_param_name_id': None,
          },
        ],
      }

    @pytest.fixture(scope='class')
    def expected_post_dataset_version_records_http(self, expected_post_dataset_version_records_https) -> Dict:
      http_endpoint_version = copy.deepcopy(expected_post_dataset_version_records_https)
      http_endpoint_version['id'] = 6
      http_endpoint_version['port'] = '80'
      http_endpoint_version['body_param_names'][0]['endpoint_id'] = 6
      http_endpoint_version['body_param_names'][1]['endpoint_id'] = 6
      http_endpoint_version['body_param_names'][2]['endpoint_id'] = 6
      http_endpoint_version['response_param_names'][0]['endpoint_id'] = 6
      return http_endpoint_version

    def test_adapt_from_file(self, open_api_endpoint_adapter, uspto_file_path, expected_get_root_https, expected_get_root_http, expected_get_dataset_version_fields_https, expected_get_dataset_version_fields_http, expected_post_dataset_version_records_https, expected_post_dataset_version_records_http):
      adapter = open_api_endpoint_adapter
      file_path = uspto_file_path 

      endpoints = adapter.adapt_from_file(file_path)

      assert len(endpoints) == 6

      get_root_https_endpoint = endpoints[0]
      get_dataset_version_fields_https = endpoints[1]
      post_dataset_version_records_https = endpoints[2]
      get_root_http_endpoint = endpoints[3]
      get_dataset_version_fields_http = endpoints[4]
      post_dataset_version_records_http = endpoints[5]

      assert get_root_https_endpoint == expected_get_root_https
      assert get_root_http_endpoint == expected_get_root_http
      assert get_dataset_version_fields_https == expected_get_dataset_version_fields_https
      assert get_dataset_version_fields_http == expected_get_dataset_version_fields_http
      assert post_dataset_version_records_https == expected_post_dataset_version_records_https
      assert post_dataset_version_records_http == expected_post_dataset_version_records_http


  class TestWhenAdaptingPetstoreReferences():

    @pytest.fixture(scope='class')
    def expected_get_v1_pets_ref(self) -> Dict:
      return {
        'id': 1,
        'method': 'POST',
        'host': 'petstore.swagger.io',
        'port': '80',
        'match_pattern': '/v1/pets',
        'path': '/v1/pets',
        'body_param_names': [
          {
            'body_param_name_id': None,
            'endpoint_id': 1,
            'id': 1,
            'inferred_type': 'String',
            'is_deterministic': True,
            'is_required': True,
            'name': 'name',
            'query': 'name',
            'values': ['']
          },
          {
            'body_param_name_id': None,
            'endpoint_id': 1,
            'id': 2,
            'inferred_type': 'String',
            'is_deterministic': True,
            'is_required': False,
            'name': 'tag',
            'query': 'tag'
          },
          {
            'body_param_name_id': None,
            'endpoint_id': 1,
            'id': 3,
            'inferred_type': 'String',
            'is_deterministic': True,
            'is_required': True,
            'name': 'color',
            'query': 'color',
            'values': ['']
          },
          {
            'body_param_name_id': None,
            'endpoint_id': 1,
            'id': 4,
            'inferred_type': 'Integer',
            'is_deterministic': True,
            'is_required': True,
            'name': 'age',
            'query': 'age',
            'values': [0]
          },
          {
            'body_param_name_id': None,
            'endpoint_id': 1,
            'id': 5,
            'inferred_type': 'Hash',
            'is_deterministic': True,
            'is_required': False,
            'name': 'adoption',
            'query': 'adoption',
          },
          {
            'body_param_name_id': 5,
            'endpoint_id': 1,
            'id': 6,
            'inferred_type': 'Boolean',
            'is_deterministic': True,
            'is_required': True,
            'name': 'adopted',
            'query': 'adoption.adopted',
            'values': [False],
          },
          {
            'body_param_name_id': 5,
            'endpoint_id': 1,
            'id': 7,
            'inferred_type': 'String',
            'is_deterministic': True,
            'is_required': False,
            'name': 'shelter',
            'query': 'adoption.shelter',
          },
          {
            'body_param_name_id': None,
            'endpoint_id': 1,
            'id': 8,
            'inferred_type': 'Integer',
            'is_deterministic': True,
            'is_required': True,
            'name': 'id',
            'query': 'id',
            'values': [0],
          },
        ],
        'path_segment_names': [
          {
            'name': 'v1',
            'type': 'static',
          },
          {
            'name': 'pets',
            'type': 'static',
          },
        ],
      }

    @pytest.fixture(scope='class')
    def expected_get_v2_pets_ref(self) -> Dict:
      return {
        'host': 'petstore.swagger.io',
        'id': 2,
        'match_pattern': '/v2/pets',
        'method': 'POST',
        'path': '/v2/pets',
        'port': '80',
        'body_param_names': [
          {
            'body_param_name_id': None,
            'endpoint_id': 2,
            'id': 1,
            'inferred_type': 'Hash',
            'is_deterministic': True,
            'is_required': False,
            'name': 'base',
            'query': 'base'
          },
          {
            'body_param_name_id': 1,
            'endpoint_id': 2,
            'id': 2,
            'inferred_type': 'String',
            'is_deterministic': True,
            'is_required': True,
            'name': 'name',
            'query': 'base.name',
            'values': ['']
          },
          {
            'body_param_name_id': 1,
            'endpoint_id': 2,
            'id': 3,
            'inferred_type': 'String',
            'is_deterministic': True,
            'is_required': False,
            'name': 'tag',
            'query': 'base.tag'
          },
          {
            'body_param_name_id': None,
            'endpoint_id': 2,
            'id': 4,
            'inferred_type': 'Hash',
            'is_deterministic': True,
            'is_required': False,
            'name': 'extra',
            'query': 'extra'
          },
          {
            'body_param_name_id': 4,
            'endpoint_id': 2,
            'id': 5,
            'inferred_type': 'String',
            'is_deterministic': True,
            'is_required': True,
            'name': 'color',
            'query': 'extra.color',
            'values': ['']
          },
          {
            'body_param_name_id': 4,
            'endpoint_id': 2,
            'id': 6,
            'inferred_type': 'Integer',
            'is_deterministic': True,
            'is_required': True,
            'name': 'age',
            'query': 'extra.age',
            'values': [0]
          },
          {
            'body_param_name_id': 4,
            'endpoint_id': 2,
            'id': 7,
            'inferred_type': 'Hash',
            'is_deterministic': True,
            'is_required': False,
            'name': 'adoption',
            'query': 'extra.adoption'
          },
          {
            'body_param_name_id': 7,
            'endpoint_id': 2,
            'id': 8,
            'inferred_type': 'Boolean',
            'is_deterministic': True,
            'is_required': True,
            'name': 'adopted',
            'query': 'extra.adoption.adopted',
            'values': [False]
          },
          {
            'body_param_name_id': 7,
            'endpoint_id': 2,
            'id': 9,
            'inferred_type': 'String',
            'is_deterministic': True,
            'is_required': False,
            'name': 'shelter',
            'query': 'extra.adoption.shelter'
          },
          {
            'body_param_name_id': None,
            'endpoint_id': 2,
            'id': 10,
            'inferred_type': 'Integer',
            'is_deterministic': True,
            'is_required': True,
            'name': 'id',
            'query': 'id',
            'values': [0]}
          ],
          'path_segment_names': [
            {
              'name': 'v2',
              'type': 'static',
            },
            {
              'name': 'pets',
              'type': 'static',
            },
          ],
        }

    def test_adapt_from_file(self, open_api_endpoint_adapter, petstore_references_file_path, expected_get_v1_pets_ref, expected_get_v2_pets_ref):
      adapter = open_api_endpoint_adapter
      file_path = petstore_references_file_path

      endpoints = adapter.adapt_from_file(file_path)

      assert len(endpoints) == 2

      get_v1_pets_ref = endpoints[0]
      get_v2_pets_ref = endpoints[1]

      assert get_v1_pets_ref == expected_get_v1_pets_ref
      assert get_v2_pets_ref == expected_get_v2_pets_ref

  class TestWhenAdaptingPetstoreSwaggerIo():

    @pytest.fixture(scope='class')
    def expected_put_v3_pets_ref(self) -> Dict:
      return {
        'host': '-',
        'id': 1,
        'match_pattern': '/v3/pet',
        'method': 'PUT',
        'path': '/v3/pet',
        'port': '0',
        'body_param_names': [
          {
            'body_param_name_id': None,
            'endpoint_id': 1,
            'id': 1,
            'inferred_type': 'Integer',
            'is_deterministic': True,
            'is_required': False,
            'name': 'id',
            'query': 'id'
          },
          {
            'body_param_name_id': None,
            'endpoint_id': 1,
            'id': 2,
            'inferred_type': 'String',
            'is_deterministic': True,
            'is_required': True,
            'name': 'name',
            'query': 'name',
            'values': ['']
          },
          {
            'body_param_name_id': None,
            'endpoint_id': 1,
            'id': 3,
            'inferred_type': 'Hash',
            'is_deterministic': True,
            'is_required': False,
            'name': 'category',
            'query': 'category'
          },
          {
            'body_param_name_id': 3,
            'endpoint_id': 1,
            'id': 4,
            'inferred_type': 'Integer',
            'is_deterministic': True,
            'is_required': False,
            'name': 'id',
            'query': 'category.id'
          },
          {
            'body_param_name_id': 3,
            'endpoint_id': 1,
            'id': 5,
            'inferred_type': 'String',
            'is_deterministic': True,
            'is_required': False,
            'name': 'name',
            'query': 'category.name',
          },
          {
            'body_param_name_id': None,
            'endpoint_id': 1,
            'id': 6,
            'inferred_type': 'Array',
            'is_deterministic': True,
            'is_required': True,
            'name': 'photoUrls',
            'query': 'photoUrls',
            'values': [[]],
          },
          {
            'body_param_name_id': 6,
            'endpoint_id': 1,
            'id': 7,
            'inferred_type': 'String',
            'is_deterministic': True,
            'is_required': False,
            'name': 'photoUrlsElement',
            'query': 'photoUrls[*]'
          },
          {
            'body_param_name_id': None,
            'endpoint_id': 1,
            'id': 8,
            'inferred_type': 'Array',
            'is_deterministic': True,
            'is_required': False,
            'name': 'tags',
            'query': 'tags'
          },
          {
            'body_param_name_id': 8,
            'endpoint_id': 1,
            'id': 9,
            'inferred_type': 'Hash',
            'is_deterministic': True,
            'is_required': False,
            'name': 'tagsElement',
            'query': 'tags[*]'
          },
          {
            'body_param_name_id': 9,
            'endpoint_id': 1,
            'id': 10,
            'inferred_type': 'Integer',
            'is_deterministic': True,
            'is_required': False,
            'name': 'id',
            'query': 'tags[*].id'
          },
          {
            'body_param_name_id': 9,
            'endpoint_id': 1,
            'id': 11,
            'inferred_type': 'String',
            'is_deterministic': True,
            'is_required': False,
            'name': 'name',
            'query': 'tags[*].name',
          },
          {
            'body_param_name_id': None,
            'endpoint_id': 1,
            'id': 12,
            'inferred_type': 'String',
            'is_deterministic': True,
            'is_required': False,
            'name': 'status',
            'query': 'status'
          },
        ],
        'path_segment_names': [
          {
              'name': 'v3',
              'type': 'static',
          },
          {
              'name': 'pet',
              'type': 'static',
          },
        ],
        'response_param_names': [
          {
              'endpoint_id': 1,
              'id': 1,
              'inferred_type': 'Integer',
              'is_deterministic': True,
              'is_required': False,
              'name': 'id',
              'query': 'id',
              'response_param_name_id': None,
          },
          {
              'endpoint_id': 1,
              'id': 2,
              'inferred_type': 'String',
              'is_deterministic': True,
              'is_required': True,
              'name': 'name',
              'query': 'name',
              'response_param_name_id': None,
              'values': [''],
          },
          {
              'endpoint_id': 1,
              'id': 3,
              'inferred_type': 'Hash',
              'is_deterministic': True,
              'is_required': False,
              'name': 'category',
              'query': 'category',
              'response_param_name_id': None,
          },
          {
              'endpoint_id': 1,
              'id': 4,
              'inferred_type': 'Integer',
              'is_deterministic': True,
              'is_required': False,
              'name': 'id',
              'query': 'category.id',
              'response_param_name_id': 3,
          },
          {
              'endpoint_id': 1,
              'id': 5,
              'inferred_type': 'String',
              'is_deterministic': True,
              'is_required': False,
              'name': 'name',
              'query': 'category.name',
              'response_param_name_id': 3,
          },
          {
              'endpoint_id': 1,
              'id': 6,
              'inferred_type': 'Array',
              'is_deterministic': True,
              'is_required': True,
              'name': 'photoUrls',
              'query': 'photoUrls',
              'response_param_name_id': None,
              'values': [[]],
          },
          {
            'endpoint_id': 1,
            'id': 7,
            'inferred_type': 'String',
            'is_deterministic': True,
            'is_required': False,
            'name': 'photoUrlsElement',
            'query': 'photoUrls[*]',
            'response_param_name_id': 6,
          },
          {
            'endpoint_id': 1,
            'id': 8,
            'inferred_type': 'Array',
            'is_deterministic': True,
            'is_required': False,
            'name': 'tags',
            'query': 'tags',
            'response_param_name_id': None,
          },
          {
            'endpoint_id': 1,
            'id': 9,
            'inferred_type': 'Hash',
            'is_deterministic': True,
            'is_required': False,
            'name': 'tagsElement',
            'query': 'tags[*]',
            'response_param_name_id': 8,
          },
          {
            'endpoint_id': 1,
            'id': 10,
            'inferred_type': 'Integer',
            'is_deterministic': True,
            'is_required': False,
            'name': 'id',
            'query': 'tags[*].id',
            'response_param_name_id': 9,
          },
          {
            'endpoint_id': 1,
            'id': 11,
            'inferred_type': 'String',
            'is_deterministic': True,
            'is_required': False,
            'name': 'name',
            'query': 'tags[*].name',
            'response_param_name_id': 9,
          },
          {
            'endpoint_id': 1,
            'id': 12,
            'inferred_type': 'String',
            'is_deterministic': True,
            'is_required': False,
            'name': 'status',
            'query': 'status',
            'response_param_name_id': None,
          },
        ],
      }

    @pytest.fixture(scope='class')
    def expected_post_v3_user_createwithlist(self) -> Dict:
      return {
        'host': '-',
        'id': 14,
        'match_pattern': '/v3/user/createWithList',
        'method': 'POST',
        'path': '/v3/user/createWithList',
        'port': '0',
        'body_param_names': [
          {
            'body_param_name_id': None,
            'endpoint_id': 14,
            'id': 1,
            'inferred_type': 'Hash',
            'is_deterministic': True,
            'is_required': False,
            'name': 'Element',
            'query': '[*]'
          },
          {
            'body_param_name_id': 1,
            'endpoint_id': 14,
            'id': 2,
            'inferred_type': 'Integer',
            'is_deterministic': True,
            'is_required': False,
            'name': 'id',
            'query': '[*].id'
          },
          {
            'body_param_name_id': 1,
            'endpoint_id': 14,
            'id': 3,
            'inferred_type': 'String',
            'is_deterministic': True,
            'is_required': False,
            'name': 'username',
            'query': '[*].username'
          },
          {
            'body_param_name_id': 1,
            'endpoint_id': 14,
            'id': 4,
            'inferred_type': 'String',
            'is_deterministic': True,
            'is_required': False,
            'name': 'firstName',
            'query': '[*].firstName'
          },
          {
            'body_param_name_id': 1,
            'endpoint_id': 14,
            'id': 5,
            'inferred_type': 'String',
            'is_deterministic': True,
            'is_required': False,
            'name': 'lastName',
            'query': '[*].lastName'
          },
          {
            'body_param_name_id': 1,
            'endpoint_id': 14,
            'id': 6,
            'inferred_type': 'String',
            'is_deterministic': True,
            'is_required': False,
            'name': 'email',
            'query': '[*].email'
          },
          {
            'body_param_name_id': 1,
            'endpoint_id': 14,
            'id': 7,
            'inferred_type': 'String',
            'is_deterministic': True,
            'is_required': False,
            'name': 'password',
            'query': '[*].password'
          },
          {
            'body_param_name_id': 1,
            'endpoint_id': 14,
            'id': 8,
            'inferred_type': 'String',
            'is_deterministic': True,
            'is_required': False,
            'name': 'phone',
            'query': '[*].phone'
          },
          {
            'body_param_name_id': 1,
            'endpoint_id': 14,
            'id': 9,
            'inferred_type': 'Integer',
            'is_deterministic': True,
            'is_required': False,
            'name': 'userStatus',
            'query': '[*].userStatus'
          }
        ],
        'path_segment_names': [
          {
            'name': 'v3',
            'type': 'static',
          },
          {
            'name': 'user',
            'type': 'static',
          },
          {
            'name': 'createWithList',
            'type': 'static',
          },
        ],
        'response_param_names': [
          {
            'endpoint_id': 14,
            'id': 1,
            'inferred_type': 'Integer',
            'is_deterministic': True,
            'is_required': False,
            'name': 'id',
            'query': 'id',
            'response_param_name_id': None,
          },
          {
            'endpoint_id': 14,
            'id': 2,
            'inferred_type': 'String',
            'is_deterministic': True,
            'is_required': False,
            'name': 'username',
            'query': 'username',
            'response_param_name_id': None,
          },
          {
            'endpoint_id': 14,
            'id': 3,
            'inferred_type': 'String',
            'is_deterministic': True,
            'is_required': False,
            'name': 'firstName',
            'query': 'firstName',
            'response_param_name_id': None,
          },
          {
            'endpoint_id': 14,
            'id': 4,
            'inferred_type': 'String',
            'is_deterministic': True,
            'is_required': False,
            'name': 'lastName',
            'query': 'lastName',
            'response_param_name_id': None,
          },
          {
            'endpoint_id': 14,
            'id': 5,
            'inferred_type': 'String',
            'is_deterministic': True,
            'is_required': False,
            'name': 'email',
            'query': 'email',
            'response_param_name_id': None,
          },
          {
            'endpoint_id': 14,
            'id': 6,
            'inferred_type': 'String',
            'is_deterministic': True,
            'is_required': False,
            'name': 'password',
            'query': 'password',
            'response_param_name_id': None,
          },
          {
            'endpoint_id': 14,
            'id': 7,
            'inferred_type': 'String',
            'is_deterministic': True,
            'is_required': False,
            'name': 'phone',
            'query': 'phone',
            'response_param_name_id': None,
          },
          {
            'endpoint_id': 14,
            'id': 8,
            'inferred_type': 'Integer',
            'is_deterministic': True,
            'is_required': False,
            'name': 'userStatus',
            'query': 'userStatus',
            'response_param_name_id': None,
          },
        ],

      }

    def test_adapt_from_file(self, open_api_endpoint_adapter, petstore_swagger_io_file_path, expected_put_v3_pets_ref, expected_post_v3_user_createwithlist):
      adapter = open_api_endpoint_adapter
      file_path = petstore_swagger_io_file_path

      endpoints = adapter.adapt_from_file(file_path)

      assert len(endpoints) == 19

      put_v3_pets_ref = endpoints[0]
      post_v3_user_createwithlist = endpoints[13]

      assert put_v3_pets_ref == expected_put_v3_pets_ref
      assert post_v3_user_createwithlist == expected_post_v3_user_createwithlist




import pdb
import requests

from stoobly_agent.app.models.adapters.python.response import PythonResponseAdapterFactory
from stoobly_agent.app.models.adapters.raw_http_response_adapter import RawHttpResponseAdapter
from stoobly_agent.app.proxy.mitmproxy.response_facade import MitmproxyResponseFacade
from stoobly_agent.app.proxy.record.response_string import ResponseString

class TestRequestString():
    
    def test_www_google_com(self):
        url='https://www.google.com'
        response = requests.get(url, stream=True)

        response_string = self.__to_response_string(response)
        raw_response_string = response_string.get()
        _response = RawHttpResponseAdapter(raw_response_string).to_response()

        self.__test_equivalence(_response, response)

    def __test_equivalence(self, response: requests.Response, _response: requests.Response):
        assert _response.raw.data == response.raw.data

        for key, val in _response.headers.items():
            assert val == response.headers.get(key)

    def __to_response_string(self, response: requests.Response):
        mitmproxy_response = PythonResponseAdapterFactory(response).mitmproxy_response()
        mitmproxy_response_facade = MitmproxyResponseFacade(mitmproxy_response)
        return ResponseString(mitmproxy_response_facade, None)


# Wraps the .config.yml file in the service folder
import hashlib
import os
import pdb

from .config import Config
from .constants import (
  SERVICE_DETACHED_ENV,
  SERVICE_HOSTNAME_ENV,
  SERVICE_PRIORITY_ENV,
  SERVICE_PORT_ENV,
  SERVICE_PROXY_MODE_ENV,
  SERVICE_SCHEME_ENV
)

class ServiceConfig(Config):

  def __init__(self, dir: str, **kwargs):
    super().__init__(dir)

    self.__detached = None
    self.__hostname = None
    self.__port = None
    self.__priority = None
    self.__proxy_mode = None
    self.__scheme = None

    self.load()

    if 'detached' in kwargs:
      self.__detached = bool(kwargs.get('detached'))
    
    if 'hostname' in kwargs:
      self.__hostname = kwargs.get('hostname')

    if 'port' in kwargs:
      self.__port = kwargs.get('port')

    if 'priority' in kwargs:
      self.__priority = kwargs.get('priority')

    if 'proxy_mode' in kwargs:
      self.__proxy_mode = kwargs.get('proxy_mode')

    if 'scheme' in kwargs:
      self.__scheme = kwargs.get('scheme')

  @property
  def detached(self) -> bool:
    return not not self.__detached

  @detached.setter
  def detached(self, v):
    self.__detached = v

  @property
  def id(self):
    return hashlib.md5(os.path.basename(self.dir).encode()).hexdigest()

  @property
  def hostname(self):
    return (self.__hostname or '').strip()

  @hostname.setter
  def hostname(self, v):
    self.__hostname = v

  @property
  def port(self):
    if not self.__port:
      if self.scheme == 'https':
        return '443' 
      elif self.scheme == 'http':
        return '80'

    return self.__port

  @port.setter
  def port(self, v):
    self.__port = v

  @property
  def priority(self):
    _priority = self.__priority
    _priority = float('inf') if _priority == None else float(_priority)

    if _priority == float('inf'):
      return _priority

    return float(_priority)

  @priority.setter
  def priority(self, v):
    self.__priority = v

  @property
  def proxy_mode(self) -> str:
    if self.__proxy_mode:
      return (self.__proxy_mode or '').strip()

    if not self.hostname:
      return ''

    _proxy_mode = f"reverse:{self.scheme}://{self.hostname}"

    if not self.port:
      return _proxy_mode

    if self.scheme == 'http' and self.port == '80': 
      return _proxy_mode

    if self.scheme == 'https' and self.port == '443':
      return _proxy_mode

    return f"{_proxy_mode}:{self.port}"

  @proxy_mode.setter
  def proxy_mode(self, v):
    self.__proxy_mode = v

  @property
  def scheme(self):
    return (self.__scheme or 'https').strip()

  @scheme.setter
  def scheme(self, v):
    self.__scheme = v

  @property
  def tls(self) -> bool:
    return self.__scheme == 'https'

  def load(self, config = None):
    config = config or self.read()

    self.detached = config.get(SERVICE_DETACHED_ENV)
    self.hostname = config.get(SERVICE_HOSTNAME_ENV)
    self.port = config.get(SERVICE_PORT_ENV)
    self.priority = config.get(SERVICE_PRIORITY_ENV)
    self.proxy_mode = config.get(SERVICE_PROXY_MODE_ENV)
    self.scheme = config.get(SERVICE_SCHEME_ENV)

  def to_dict(self):
    return {
      'detached': self.detached,
      'hostname': self.hostname,
      'port': self.port,
      'priority': self.priority,
      'proxy_mode': self.proxy_mode,
      'scheme': self.scheme if self.hostname else '',
    }

  def write(self):
    config = {}

    if self.hostname:
      config[SERVICE_HOSTNAME_ENV] = self.hostname

    if self.port:
      config[SERVICE_PORT_ENV] = self.port

    if self.priority:
      config[SERVICE_PRIORITY_ENV] = self.priority

    if self.scheme:
      config[SERVICE_SCHEME_ENV] = self.scheme

    config[SERVICE_DETACHED_ENV] = bool(self.detached)

    config[SERVICE_PROXY_MODE_ENV] = self.proxy_mode

    super().write(config)


import pdb
import requests

from .mitmproxy_adapter import MitmproxyResponseAdapter
from .python_adapter import PythonResponseAdapter

class OrmResponseAdapterFactory():

  def __init__(self, response: requests.Response):
    self.__response = response

  def mitmproxy_response(self):
    return MitmproxyResponseAdapter(self.__response).adapt()

  def python_response(self) -> requests.Response:
    return PythonResponseAdapter(self.__response).adapt()


import requests
import urllib
import pdb

from stoobly_agent.app.models.types import HeaderNameCreateParams

from ..logger import Logger
from .endpoints_resource import EndpointsResource

LOG_ID = 'ResponseHeaderNamesResource'

class ResponseHeaderNamesResource(EndpointsResource):
  RESPONSE_HEADER_NAME_ENDPOINT = 'response_header_names'

  def create(self, endpoint_id: int, params: HeaderNameCreateParams = {}):
    url = f"{self.service_url}/{self.ENDPOINTS_ENDPOINT}/{endpoint_id}/{self.RESPONSE_HEADER_NAME_ENDPOINT}"
    return self.post(url, headers=self.default_headers, json=params)

  def index(self, endpoint_id: int, query_params = {}) -> requests.Response:
    url = f"{self.service_url}/{self.ENDPOINTS_ENDPOINT}/{endpoint_id}/{self.RESPONSE_HEADER_NAME_ENDPOINT}"

    Logger.instance(LOG_ID).debug(f"{url}?{urllib.parse.urlencode(query_params)}")

    return self.get(url, headers=self.default_headers, params=query_params)

  def show(self, endpoint_id: int, response_header_name_id: int, query_params = {}) -> requests.Response:
    url = f"{self.service_url}/{self.ENDPOINTS_ENDPOINT}/{endpoint_id}/{self.RESPONSE_HEADER_NAME_ENDPOINT}/{response_header_name_id}"

    Logger.instance(LOG_ID).debug(f"{url}?{urllib.parse.urlencode(query_params)}")

    return self.get(url, headers=self.default_headers, params=query_params)
  
  def destroy(self, endpoint_id: int, response_header_name_id: int, query_params = {}) -> requests.Response:
    url = f"{self.service_url}/{self.ENDPOINTS_ENDPOINT}/{endpoint_id}/{self.RESPONSE_HEADER_NAME_ENDPOINT}/{response_header_name_id}"

    Logger.instance(LOG_ID).debug(f"{url}?{urllib.parse.urlencode(query_params)}")

    return self.delete(url, headers=self.default_headers, params=query_params)


import click
import sys

from stoobly_agent.app.settings import Settings

from .helpers.print_service import FORMATS, print_reports, select_print_options
from .helpers.report_facade import ReportFacade
from .helpers.validations import *

@click.group(
    epilog="Run 'stoobly-agent report COMMAND --help' for more information on a command.",
    help="Manage test reports"
)
@click.pass_context
def report(ctx):
    pass

@report.command(
    help="Create a report"
)
@click.option('--description', help='Report description.')
@click.option('--format', type=click.Choice(FORMATS), help='Format output.')
@click.option('--project-key', help='Project to create report in.')
@click.option('--select', multiple=True, help='Select column(s) to display.')
@click.option('--without-headers', is_flag=True, default=True, help='Disable printing column headers.')
@click.argument('name')
def create(**kwargs):
    print_options = select_print_options(kwargs)
    
    settings = Settings.instance()
    project_key = resolve_project_key_and_validate(kwargs, settings)

    if not project_key:
        project_key = settings.proxy.intercept.project_key

    report = ReportFacade(settings)

    try:
        res = report.create(project_key, kwargs['name'], kwargs['description'])
    except AssertionError as e:
        return print(e, file=sys.stderr)

    print_reports([res], **print_options)

@report.command(
    help="Show created reports"
)
@click.option('--format', type=click.Choice(FORMATS), help='Format output.')
@click.option('--page', default=0)
@click.option('--project-key', help='Project to create scenario in.')
@click.option('--select', multiple=True, help='Select column(s) to display.')
@click.option('--sort-by', default='created_at', help='created_at|name')
@click.option('--sort-order', default='desc', help='asc | desc')
@click.option('--size', default=10)
@click.option('--without-headers', is_flag=True, default=False, help='Disable printing column headers.')
def list(**kwargs):
    print_options = select_print_options(kwargs)

    settings = Settings.instance()
    project_key = resolve_project_key_and_validate(kwargs, settings)
    del kwargs['project_key']

    report = ReportFacade(Settings.instance())

    try:
        reports_response = report.index(project_key, kwargs)
    except AssertionError as e:
        return print(e, file=sys.stderr)

    if len(reports_response['list']) == 0:
        print('No reports found.', file=sys.stderr)
    else:
        print_reports(reports_response['list'], **print_options)


import pdb

from mitmproxy.http import Headers, Response as MitmproxyResponse
from mitmproxy.coretypes import multidict
from typing import Callable, List

from stoobly_agent.app.settings.constants import request_component
from stoobly_agent.app.settings.rewrite_rule import ParameterRule, RewriteRule
from stoobly_agent.config.constants import custom_headers
from stoobly_agent.lib.logger import Logger, bcolors
from stoobly_agent.lib.utils import jmespath
from stoobly_agent.lib.utils.decode import decode

from .request_facade import MitmproxyRequestFacade
from .response_body_facade import MitmproxyResponseBodyFacade
from .response import Response

LOG_ID = 'Response'

class MitmproxyResponseFacade(Response):

    def __init__(self, response: MitmproxyResponse):
        self.response = response

        self.__body = MitmproxyResponseBodyFacade(response)
        self.__parameter_rules: List[ParameterRule] = []

    @property
    def code(self):
        return self.response.status_code

    @property
    def content_type(self):
        return self.headers.get('content-type')

    @property
    def headers(self):
        return self.__filter_custom_headers(self.response.headers)

    @property
    def body(self):
        return self.response.raw_content or ''

    @property
    def http_verison(self):
        return self.response.http_version

    def decode_body(self):
        # Decodes content (if Content-Encoding header is set)
        self.content = self.response.content

        # Update Content-Lenght header to decoded content length
        self.response.headers['content-length'] = str(len(self.content))

    def with_parameter_rules(self, rules: List[RewriteRule], request_facade: MitmproxyRequestFacade):
        if type(rules) == list:
            self.__parameter_rules = request_facade.select_parameter_rules(rules)
        return self 

    def rewrite(self):
        rewrites = self.__parameter_rules

        if len(rewrites) != 0:
            self.__rewrite_headers(rewrites)
            self.__rewrite_content(rewrites)
    
    def __apply_rewrites(self, params: dict, rewrites: List[ParameterRule], handler: Callable):
        if len(rewrites) == 0:
            return

        for rewrite in rewrites:
            jmespath.search(rewrite.name, params, {
                'replacements': [handler(rewrite) if handler else rewrite.value],
            })

    def __rewrite_handler(self, rewrite: ParameterRule) -> str:
        Logger.instance(LOG_ID).info(f"{bcolors.OKCYAN}Rewriting{bcolors.ENDC} {rewrite.type.lower()} {rewrite.name} => {rewrite.value}")
        return rewrite.value

    def __rewrite_headers(self, rewrites: List[ParameterRule]):
        self.__apply_headers(rewrites, self.__rewrite_handler)

    def __rewrite_content(self, rewrites: List[ParameterRule]):
        self.__apply_content(rewrites, self.__rewrite_handler)

    def __apply_headers(self, rewrites: List[ParameterRule], handler: Callable):
        rewrites = list(filter(lambda rewrite: rewrite.type == request_component.RESPONSE_HEADER, rewrites))
        self.__apply_rewrites(self.response.headers, rewrites, handler)

    def __apply_content(self, rewrites: List[ParameterRule], handler: Callable):
        rewrites = list(filter(lambda rewrite: rewrite.type == request_component.RESPONSE_PARAM, rewrites))
        if len(rewrites) == 0:
            return

        content_type = self.content_type
        parsed_content = self.__body.get(content_type)

        if not self.__is_iterable(parsed_content):
            return 

        self.__apply_rewrites(parsed_content, rewrites, handler)
        self.__body.set(parsed_content, content_type)

    def __filter_custom_headers(self, response_headers: Headers):
        '''
        Remove custom headers
        '''
        _response_headers = Headers(**response_headers)

        headers = custom_headers.__dict__
        for key in headers:
            if key[0:2] == '__' and key[-2:] == '__':
                continue

            name = headers[key]

            if name not in response_headers:
                continue

            _response_headers.pop(name)

        return _response_headers

    def __is_iterable(self, v):
        return isinstance(v, dict) or isinstance(v, multidict.MultiDictView) or isinstance(v, list)


import pdb
from typing import List, TypedDict

from stoobly_agent.lib.orm.trace import Trace

class Alias(TypedDict):
  name: str
  value: str

def parse_aliases(aliases: List[str]) -> List[Alias]: 
  def parse_alias_string(_alias):
    toks = _alias.split('=', 1)

    try:
      value = eval(toks[1])
    except Exception as e:
      if len(toks) > 1:
        value = toks[1]
      else:
        value = None

    return {
      'name': toks[0],
      'value': value,
    }

  return list(map(parse_alias_string, aliases))

def adapt_trace_aliases(trace_id: int) -> List[Alias]:
  trace = Trace.find_by(id=trace_id)
  if not trace:
    return []

  trace_aliases = trace.trace_aliases

  def trace_alias_to_alias(trace_alias):
    return {
      'name': trace_alias.name,
      'value': trace_alias.value,
    }
    
  return list(map(trace_alias_to_alias, trace_aliases))


import os
import pdb
import requests
import time

from mitmproxy.http import HTTPFlow as MitmproxyHTTPFlow, Request as MitmproxyRequest
from typing import Callable, TypedDict

from stoobly_agent.app.models.request_model import RequestModel
from stoobly_agent.app.proxy.mitmproxy.request_facade import MitmproxyRequestFacade
from stoobly_agent.app.proxy.utils.rewrite_rules_to_ignored_components_service import rewrite_rules_to_ignored_components
from stoobly_agent.config.constants import custom_headers, env_vars, lifecycle_hooks, mock_policy, request_origin
from stoobly_agent.lib.logger import bcolors, Logger

from .constants import custom_response_codes
from .mock.context import MockContext
from .mock.eval_fixtures_service import eval_fixtures
from .mock.eval_request_service import inject_eval_request
from .utils.allowed_request_service import get_active_mode_policy
from .utils.request_handler import reverse_proxy
from .utils.response_handler import bad_request, enable_cors, pass_on
from .utils.rewrite import rewrite_request, rewrite_response

LOG_ID = 'Mock'

class MockOptions(TypedDict):
    failure: Callable
    ignored_components: list 
    infer: bool
    no_rewrite: bool
    success: Callable

def handle_request_mock_generic_without_rewrite(context: MockContext, **options: MockOptions):
    options['no_rewrite'] = True
    handle_request_mock_generic(context, **options)

###
#
# 1. Rewrites mock request by default
# 2. BEFORE_MOCK gets triggered
# 3. AFTER_MOCK gets triggered
#
# @param request [mitmproxy.http.Request]
# @param settings [Dict]
#
def handle_request_mock_generic(context: MockContext, **options: MockOptions):
    handle_error = options['error'] if 'error' in options and callable(options['error']) else None
    handle_failure = options['failure'] if 'failure' in options and callable(options['failure']) else None
    handle_success = options['success'] if 'success' in options and callable(options['success']) else None
    intercept_settings = context.intercept_settings
    request: MitmproxyRequest = context.flow.request
    res = None

    policy = get_active_mode_policy(request, intercept_settings)
    if policy == mock_policy.NONE:
        if handle_error:
            res = handle_error(context)

        return pass_on(context.flow, res)

    if policy not in [mock_policy.ALL, mock_policy.FOUND]:
        if handle_error:
            res = handle_error(context)

        return bad_request(
            context.flow,
            "Valid env MOCK_POLICY: %s, Got: %s" %
            ([mock_policy.ALL, mock_policy.FOUND, mock_policy.NONE], policy)
        )

    if not options.get('no_rewrite'):
        __rewrite_request(context)

    __mock_hook(lifecycle_hooks.BEFORE_MOCK, context)

    # If ignore rules are set, then ignore specified request parameters
    ignore_rules = intercept_settings.ignore_rules
    if len(ignore_rules) > 0:
        request_facade = MitmproxyRequestFacade(request)
        _ignore_rules = request_facade.select_parameter_rules(ignore_rules)
        ignored_components = rewrite_rules_to_ignored_components(_ignore_rules)
        options['ignored_components'] += ignored_components if 'ignored_components' in options else ignored_components 

    request_model = RequestModel(intercept_settings.settings)
    eval_request = inject_eval_request(request_model, intercept_settings)

    if policy == mock_policy.ALL:
        res = eval_request_with_retry(context, eval_request, **options) 

        context.with_response(res)
    elif policy == mock_policy.FOUND:
        res = eval_request_with_retry(context, eval_request, **options) 

        context.with_response(res)

        if res.status_code in [custom_response_codes.NOT_FOUND, custom_response_codes.IGNORE_COMPONENTS]:
            try:
                return __handle_found_policy(context) # Continue proxying the request
            except RuntimeError:
                # Do nothing, return custom error response
                pass
    
    if res.status_code == custom_response_codes.NOT_FOUND:
        if handle_failure:
            res = handle_failure(context) or res
    else:
        if handle_success:
            res = handle_success(context) or res

    return pass_on(context.flow, res)

def eval_request_with_retry(context: MockContext, eval_request, **options: MockOptions):
    request = context.flow.request
    infer = bool(options.get('infer'))
    ignored_components = options['ignored_components'] if 'ignored_components' in options else []

    res: requests.Response = eval_request(request, ignored_components)

    if res.status_code == custom_response_codes.IGNORE_COMPONENTS:
        ignored_components.append(res.content)
        res = eval_request(request, ignored_components, infer=infer, retry=1)

    if res.status_code == custom_response_codes.NOT_FOUND:
        intercept_settings = context.intercept_settings
        fixture = eval_fixtures(
            request,
            public_directory_path=intercept_settings.public_directory_path,
            response_fixtures=intercept_settings.response_fixtures
        )
        if fixture:
            res = fixture

    return res

def handle_request_mock(context: MockContext):
    handle_request_mock_generic(
        context,
        failure=lambda context: __handle_mock_failure(context),
        success=lambda context: __handle_mock_success(context)
    )

###
#
# 1. Rewrites mock response (if mock found)
# 2. AFTER_MOCK gets triggered (if mock found)
#
def handle_response_mock(context: MockContext):
    response = context.flow.response
    request_key = response.headers.get(custom_headers.MOCK_REQUEST_KEY)

    if request_key:
        request = context.flow.request
        Logger.instance(LOG_ID).info(f"{bcolors.OKBLUE}Mocked{bcolors.ENDC} {request.url} -> {request_key}")

        __rewrite_response(context)
        __mock_hook(lifecycle_hooks.AFTER_MOCK, context)

def __handle_mock_failure(context: MockContext) -> None:
    flow = context.flow
    request = flow.request

    if request.method.upper() != 'OPTIONS':
        return False
        
    # Default OPTIONS request to allow CORS
    enable_cors(flow)

    return True

def __handle_found_policy(context: MockContext) -> None:
    req = context.flow.request
    intercept_settings = context.intercept_settings
    upstream_url = intercept_settings.upstream_url

    if req.headers.get(custom_headers.REQUEST_ORIGIN) == request_origin.PROXY:
        # If this header is set, then it is likely that we are going to infinite loop
        # Unless we stop sending the same request
        raise RuntimeError(f"Request originated from {request_origin.PROXY}")
    else:
        req.headers[custom_headers.REQUEST_ORIGIN] = request_origin.PROXY

    Logger.instance(LOG_ID).debug(f"UpstreamUrl: {upstream_url}")

    reverse_proxy(req, upstream_url, {})

def __handle_mock_success(context: MockContext) -> None:
    if os.environ.get(env_vars.AGENT_SIMULATE_LATENCY):
        response = context.response
        start_time = context.start_time
        __simulate_latency(response.headers.get(custom_headers.RESPONSE_LATENCY), start_time)

def __rewrite_request(context: MockContext):
    # Rewrite request with paramter rules for mock

    intercept_settings = context.intercept_settings
    rewrite_rules = intercept_settings.mock_rewrite_rules

    if len(rewrite_rules) > 0:
        rewrite_request(context.flow, rewrite_rules) 

def __rewrite_response(context: MockContext):
    # Rewrite request with paramter rules for mock

    intercept_settings = context.intercept_settings
    rewrite_rules = intercept_settings.mock_rewrite_rules

    if len(rewrite_rules) > 0:
        rewrite_response(context.flow, rewrite_rules) 

###
#
# Try to simulate expected response latency
#
# wait_time (seconds) = expected_latency - estimated_rtt_network_latency - api_latency
#
# expected_latency = provided value
# estimated_rtt_network_latency = 15ms
# api_latency = current_time - start_time of this request
#
def __simulate_latency(expected_latency: str, start_time: float) -> float:
    if not expected_latency:
        return 0

    estimated_rtt_network_latency = 0.015 # seconds
    api_latency = (time.time() - start_time)
    expected_latency = float(expected_latency) / 1000

    wait_time = expected_latency - estimated_rtt_network_latency - api_latency

    Logger.instance(LOG_ID).debug(f"Expected latency: {expected_latency}")
    Logger.instance(LOG_ID).debug(f"API latency: {api_latency}")
    Logger.instance(LOG_ID).debug(f"Wait time: {wait_time}")

    if wait_time > 0:
        time.sleep(wait_time)

    return wait_time

def __mock_hook(hook: str, context: MockContext):
    intercept_settings = context.intercept_settings
    lifecycle_hooks_module = intercept_settings.lifecycle_hooks

    if hook in lifecycle_hooks_module:
        lifecycle_hooks_module[hook](context)


import io
import mimetypes
import re
import pdb

from multipart import MultipartParser
from urllib.parse import quote

from mitmproxy.net.http import headers

CRLF = b'\r\n'

def encode(head, l):

    k = head.get("content-type")
    if k:
        k = headers.parse_content_type(k)
        if k is not None:
            try:
                boundary = k[2]["boundary"].encode("ascii")
                boundary = quote(boundary)
            except (KeyError, UnicodeError):
                return b""

            hdrs = []
            for key, value in l:
                file_type = mimetypes.guess_type(str(key))[0] or "text/plain; charset=utf-8"

                if key:
                    hdrs.append(b"--%b" % boundary.encode('utf-8'))
                    disposition = b'form-data; name="%b"' % key
                    hdrs.append(b"Content-Disposition: %b" % disposition)
                    hdrs.append(b"Content-Type: %b" % file_type.encode('utf-8'))
                    hdrs.append(b'')
                    hdrs.append(value)
                #hdrs.append(b'')

                if value is not None:
                    # If boundary is found in value then raise ValueError
                    if re.search(rb"^--%b$" % re.escape(boundary.encode('utf-8')), value):
                        raise ValueError(b"boundary found in encoded string")

            hdrs.append(b"--%b--\r\n" % boundary.encode('utf-8'))

            temp = b"\r\n".join(hdrs)
            return temp


def decode(hdrs, content):
    """
        Takes a multipart boundary encoded string and returns list of (key, value) tuples.
    """
    if not isinstance(content, bytes) and not isinstance(content, str):
        return

    v = hdrs.get("content-type")
    if not v:
        return

    v = headers.parse_content_type(v)
    if not v:
        return

    try:
        boundary = v[2]["boundary"].encode("ascii")
    except (KeyError, UnicodeError):
        return

    r = []
    parser = MultipartParser(io.BytesIO(content), boundary=boundary)
    for part in parser.parts():
        if part.content_type.lower() == 'application/octet-stream':
            r.append((part.name, part.raw))
        else:
            r.append((part.name, part.value))

        # Free up resources after use
        part.close()
    return r



import pdb
import requests

from mitmproxy.http import HTTPFlow as MitmproxyHTTPFlow
from typing import Tuple

from stoobly_agent.app.models.adapters.python import PythonRequestAdapterFactory
from stoobly_agent.app.models.helpers.create_request_params_service import build_params
from stoobly_agent.app.models.types import RequestCreateParams, RequestDestroyParams, RequestDestroyAllParams, RequestIndexSimilarParams, RequestShowParams
from stoobly_agent.app.proxy.mock.custom_not_found_response_builder import CustomNotFoundResponseBuilder
from stoobly_agent.app.proxy.mock.ignored_components_response_builder import IgnoreComponentsResponseBuilder
from stoobly_agent.app.proxy.record.joined_request import JoinedRequest
from stoobly_agent.config.constants import custom_headers, query_params as request_query_params
from stoobly_agent.lib.orm import ORM
from stoobly_agent.lib.orm.request import Request
from stoobly_agent.lib.orm.response import Response
from stoobly_agent.lib.orm.scenario import Scenario
from stoobly_agent.lib.orm.transformers.orm_to_stoobly_request_transformer import ORMToStooblyRequestTransformer
from stoobly_agent.lib.orm.types.request_columns import RequestColumns
from stoobly_agent.lib.orm.transformers import ORMToRequestTransformer, ORMToRequestsResponseTransformer
from stoobly_agent.lib.api.interfaces import RequestsIndexQueryParams, RequestsIndexResponse, RequestShowResponse

from .helpers.create_request_columns_service import build_request_columns, build_response_columns
from .helpers.search import search_request
from .helpers.snapshot_service import snapshot_request
from .helpers.tiebreak_scenario_request import access_request, generate_session_id, tiebreak_scenario_request
from .local_db_adapter import LocalDBAdapter
from .orm_request_builder import ORMRequestBuilder
from .response_adapter import LocalDBResponseAdapter

class LocalDBRequestAdapter(LocalDBAdapter):
  __request_orm = None
  __response_orm = None

  def __init__(self, request_orm: Request.__class__ = Request, response_orm: Response.__class__ = Response, scenario_orm: Scenario.__class__ = Scenario):
    self.__request_orm: Request = request_orm
    self.__response_orm: Response = response_orm
    self.__scenario_orm: Scenario = scenario_orm

  def create(self, **params: RequestCreateParams) -> Tuple[RequestShowResponse, int]:
    self.__adapt_scenario_id(params)

    flow: MitmproxyHTTPFlow = params['flow']
    joined_request: JoinedRequest = params['joined_request']
    scenario_id = params.get('scenario_id')
    uuid = params.get('uuid')

    request_columns = build_request_columns(flow, joined_request, is_deleted=False, scenario_id=scenario_id, uuid=uuid)

    with ORM.instance().db.transaction():
      request_record = self.__request_orm.create(**request_columns)

      response_columns = {
        'request_id': request_record.id,
        **build_response_columns(flow, joined_request),
      }

      self.__response_orm.create(**response_columns)

      return self.success({
        'list': [ORMToStooblyRequestTransformer(request_record, {}).transform()],
        'total': 1,
      })

  def show(self, request_id: str, **options: RequestShowParams) -> Tuple[RequestShowResponse, int]:
    request = self.__request(request_id)

    if not request:
      return self.__request_not_found()

    return self.success(ORMToStooblyRequestTransformer(request, options).transform())

  def response(self, **query_params: RequestColumns) -> requests.Response:
    self.__adapt_scenario_id(query_params)

    request = None

    if not query_params.get('request_id'):
      request_columns = { 'is_deleted': False, **query_params }

      self.__filter_request_response_columns(request_columns)

      # Find most recent matching record
      requests = self.__request_orm.where_for(**request_columns).get()

      if 'scenario_id' in query_params:
        if len(requests) > 1:
          session_id = query_params.get(request_query_params.SESSION_ID)
          request_session_id_components = { **request_columns }

          if session_id:
            request_session_id_components[request_query_params.SESSION_ID] = session_id

          # When multiple requests are matched for a scenario, return them in sequence
          request_session_id = generate_session_id(request_session_id_components)
          request = tiebreak_scenario_request(request_session_id, requests)
          access_request(request_session_id, request.id)
        else:
          request = requests.last()
      else:
        request = requests.last()
    else:
      request = self.__request_orm.find(query_params.get('request_id'))
      
      if request and request.is_deleted:
        request = None

    if not request:
      endpoint_promise = query_params.get(request_query_params.ENDPOINT_PROMISE)
      return self.__handle_request_not_found(endpoint_promise) 

    response_record = request.response
    if not response_record:
      return CustomNotFoundResponseBuilder().build()

    headers = {}
    headers[custom_headers.MOCK_REQUEST_ID] = str(request.id)
    headers[custom_headers.MOCK_REQUEST_KEY] = request.key()
    headers[custom_headers.RESPONSE_LATENCY] = str(request.latency)

    return (
      ORMToRequestsResponseTransformer(response_record)
        .with_headers(headers)
        .transform()
    )

  def __handle_request_not_found(self, endpoint_promise):
    if endpoint_promise:
      ignored_components = self.__ignored_components(endpoint_promise)

      if ignored_components:
        return IgnoreComponentsResponseBuilder().build(ignored_components)

    return CustomNotFoundResponseBuilder().build()

  def index(self, **query_params: RequestsIndexQueryParams) -> Tuple[RequestsIndexResponse, int]:
    self.__adapt_scenario_id(query_params)

    scenario_id = query_params.get('scenario_id')
    page = int(query_params.get('page') or 0)
    query = query_params.get('q')
    size = int(query_params.get('size') or 20)
    sort_by = query_params.get('sort_by') or 'id'
    sort_order = query_params.get('sort_order') or 'desc'

    is_deleted = query_params.get('filter') == 'is_deleted'
    starred = query_params.get('filter') == 'starred'
    unassigned = query_params.get('filter') == 'unassigned'

    requests = Request.where('is_deleted', is_deleted)

    if unassigned:
      requests = requests.where('scenario_id', None)
    elif scenario_id:
      requests = requests.where('scenario_id', scenario_id)
      sort_order = query_params.get('sort_order') or 'asc'

    if starred:
      requests = requests.where('starred', starred)

    if query:
      requests = search_request(requests, query)

    total = requests.count()
    requests = requests.offset(page * size).limit(size).order_by(sort_by, sort_order).get()

    return self.success({
      'list': list(map(lambda request: ORMToStooblyRequestTransformer(request, {}).transform(), requests)),
      'total': total,
    })

  def update(self, request_id: int,  **params: RequestShowResponse):
    request = self.__request(request_id)

    if not request:
      return self.__request_not_found()

    transformer = ORMToRequestTransformer(request)

    if params.get('request'):
      create_params = build_params(params['request'])
      del params['request']

      if not create_params:
        return self.bad_request('Could not parse request')

      if request.update(
        **build_request_columns(create_params['flow'], create_params['joined_request'], **params),
      ):
        response = request.response
        if response.update(**{
          **build_response_columns(create_params['flow'], create_params['joined_request']),
        }):
          return self.success(ORMToStooblyRequestTransformer(request, {}).transform())
    else:
      if params.get('method'):
        transformer.with_method(params['method'])

      if params.get('url'):
        transformer.with_url(params['url'])
        del params['url']

      if params.get('headers'):
        transformer.with_headers(params['headers'])
        del params['headers']

      if params.get('body'):
        transformer.with_body(params['body'])
        del params['body']

      if transformer.dirty:
        python_request = transformer.transform()
        adapter_factory = PythonRequestAdapterFactory(python_request)

        builder = ORMRequestBuilder()
        columns = builder.columns_from_mitmproxy_request(adapter_factory.mitmproxy_request())
        params = {
          'raw': adapter_factory.raw_request(),
          **params,
          **columns,
        }

      # Some params need to be reflected in response record
      response_params = {}

      if params.get('latency'):
        latency = int(params['latency'])
        latency = latency = 0 if latency < 0 else latency
        params['latency'] = latency
        response_params['latency'] = latency

      if params.get('status'):
        response_params['status'] = params['status']

      if request.update(params):
        if len(response_params.keys()) != 0:
          response = request.response
          LocalDBResponseAdapter(self.__request_orm).update(response.id, **response_params)

        return self.success(ORMToStooblyRequestTransformer(request, {}).transform())

    return self.internal_error('Could not update request')

  def destroy(self, request_id: int, **params: RequestDestroyParams):
    request = self.__request(request_id)

    if not request:
      return self.__request_not_found()

    if params.get('force') or request.is_deleted:
      request.delete()
    else:
      request.update({'is_deleted': True})

    return self.success(ORMToStooblyRequestTransformer(request, {}).transform())

  def destroy_all(self, **params: RequestDestroyAllParams) -> Tuple[RequestsIndexQueryParams, int]:
    self.__adapt_scenario_id(params)

    ids = params.get('ids')
    scenario_id = params.get('scenario_id')

    if not isinstance(ids, list) and not scenario_id:
      return self.bad_request('Missing ids')

    requests_query_builder = self.__request_orm

    if isinstance(ids, list):
      requests_query_builder = requests_query_builder.where_in('id', ids)

    if scenario_id:
      requests_query_builder = requests_query_builder.where_for(scenario_id=scenario_id)

    requests = requests_query_builder.get()

    res = self.success({
      'list': list(map(lambda request: ORMToStooblyRequestTransformer(request, {}).transform(), requests)),
      'total': requests.count(),
    })

    for request in requests:
      request.delete()

    return res

  def snapshot(self, request_id: str, **params):
    request = self.__request(request_id)

    if not request:
      return self.__request_not_found()

    file_path = snapshot_request(request, **params)
    if not file_path:
      return self.internal_error()

    return self.success(file_path)

  def index_similar(self, params: RequestIndexSimilarParams):
    pattern = f"%{params['pattern']}"
    candidates = None

    if params['host'] == '%':
      candidates = self.__request_orm.where('host', 'like', params['host'])
    else:
      candidates = self.__request_orm.where('host', params['host'])

    if params['port'] == '%':
      candidates = candidates.where('port', 'like', params['port'])
    else:
      candidates = candidates.where('port', params['port'])

    candidates = candidates.where('method', params['method'])
    candidates = candidates.where('path', 'like', pattern)
    if params.get('scenario_id'):
      candidates = candidates.where('scenario_id', params['scenario_id'])

    return candidates.get()

  def __filter_request_response_columns(self, request_columns: RequestCreateParams):
    if request_columns.get(request_query_params.ENDPOINT_PROMISE):
      del request_columns[request_query_params.ENDPOINT_PROMISE]

    if request_columns.get('infer'):
      del request_columns['infer']

    if request_columns.get('project_id'):
      del request_columns['project_id']

    if request_columns.get('retry'):
      del request_columns['retry']

    if request_columns.get(request_query_params.SESSION_ID):
      del request_columns[request_query_params.SESSION_ID]

  def __request(self, request_id: str):
    if self.validate_uuid(request_id):
      return self.__request_orm.find_by(uuid=request_id)
    else:
      return self.__request_orm.find(request_id)

  def __request_not_found(self):
    return self.not_found('Request not found')

  def __adapt_scenario_id(self, params: dict):
    if not params.get('scenario_id'):
      return

    scenario_id = params['scenario_id']
    if not self.validate_uuid(scenario_id):
      return

    scenario = self.__scenario_orm.find_by(uuid=scenario_id)
    if scenario:
      params['scenario_id'] = scenario.id

  def __ignored_components(self, endpoint_promise):
    endpoint = endpoint_promise()

    if endpoint:
      ignored_components = endpoint.get('ignored_components')

      if ignored_components:
        return ignored_components
    return []


import uuid

from stoobly_orator.migrations import Migration

from stoobly_agent.lib.orm.scenario import Scenario

class AddUuidColumnToScenarios(Migration):

    def up(self):
        """
        Run the migrations.
        """
        with self.schema.table('scenarios') as table:
            table.string('uuid', 36).default('').index().unique()

        for scenario in Scenario.all():
            scenario.update(uuid=str(uuid.uuid4()))

    def down(self):
        """
        Revert the migrations.
        """
        with self.schema.table('scenarios') as table:
            table.drop_column('uuid')



APP_EGRESS_NETWORK = '${APP_NETWORK}.egress'
APP_EGRESS_NETWORK_NAME = 'app.egress'
APP_EGRESS_NETWORK_TEMPLATE = '{network}.egress'
APP_INGRESS_NETWORK = '${APP_NETWORK}.ingress'
APP_INGRESS_NETWORK_NAME = 'app.ingress'
APP_INGRESS_NETWORK_TEMPLATE = '{network}.ingress'
DOCKER_COMPOSE_BASE = '.docker-compose.base.yml'
DOCKER_COMPOSE_BASE_TEMPLATE = '.docker-compose.base.template.yml'
DOCKER_COMPOSE_CUSTOM = 'docker-compose.yml'
DOCKER_COMPOSE_NETWORKS = '.docker-compose.networks.yml'
DOCKERFILE_CONTEXT = '.Dockerfile.context'
DOCKERFILE_SERVICE = 'Dockerfile.source'

# TODO: add scaffold container name templates here

# Example:
# COMPOSE_TEMPLATE = 'docker-compose.{workflow}.yml'




import os
import pdb
import time
import yaml

from shutil import copyfile
from watchdog.observers import Observer
from watchdog.events import PatternMatchingEventHandler
from yamale import *

from stoobly_agent.config.constants import env_vars, statuses
from stoobly_agent.config.data_dir import DataDir
from stoobly_agent.config.source_dir import SourceDir
from stoobly_agent.lib.logger import Logger

from .cli_settings import CLISettings
from .proxy_settings import ProxySettings
from .remote_settings import RemoteSettings
from .ui_settings import UISettings

LOG_ID = 'Settings'

class Settings:
    _instances = None

    __data_dir: DataDir = None

    __cli_settings = None
    __proxy_settings = None
    __remote_settings = None
    __ui_settings = None

    __settings_file_path = None
    __schema_file_path = None

    __load_lock = False
    __watching = False

    def __init__(self, data_dir_path: str = None):
        if Settings._instances.get(data_dir_path):
            raise RuntimeError('Call instance() instead')

        self.__data_dir = DataDir.instance(data_dir_path)
        self.__detect_paths()

        # If the config does not exist, use template
        if not os.path.exists(self.__settings_file_path):
            self.__create_default_file()

        self.__load_settings()

    @classmethod
    def instance(cls, data_dir_path: str = None):
        if not cls._instances:
            cls._instances = {}

        if not cls._instances.get(data_dir_path):
            cls._instances[data_dir_path] = cls(data_dir_path)

        return cls._instances[data_dir_path]

    @classmethod
    def handle_chdir(cls):
        '''
        Reloads data dir to be relative to new working directory
        '''
        DataDir.handle_chdir()

        if cls._instances and None in cls._instances:
            del cls._instances[None]

        return cls.instance()

    ### Statuses

    def is_debug(self):
        return os.environ.get(env_vars.LOG_LEVEL) == 'debug'

    ### Properties

    @property
    def cli(self):
        return self.__cli_settings

    @property
    def ui(self):
        return self.__ui_settings

    @property
    def path(self):
        return self.__settings_file_path

    @property
    def proxy(self):
        return self.__proxy_settings

    @property
    def remote(self):
        return self.__remote_settings

    def watch(self):
        if self.__watching:
            return False

        patterns = ['settings.yml']
        ignore_patterns = None
        ignore_directories = False
        case_sensitive = True
        event_handler = PatternMatchingEventHandler(
            patterns=patterns, 
            ignore_patterns=ignore_patterns, 
            ignore_directories=ignore_directories, 
            case_sensitive=case_sensitive
        )
        event_handler.on_modified = self.__reload_settings

        observer = Observer()
        watch_dir = os.path.dirname(self.__settings_file_path)

        observer.schedule(event_handler, watch_dir)
        observer.start()

        self.__watching = True

        return True

    ### Get

    def to_dict(self):
        if not self.__settings:
            self.__load_settings()

        return {
            **self.__settings,
            **{ 'cli': self.__cli_settings.to_dict() },
            **{ 'proxy': self.__proxy_settings.to_dict() },
            **{ 'remote': self.__remote_settings.to_dict() },
            **{ 'ui': self.__ui_settings.to_dict() },
        }

    def read(self):
        with open(self.__settings_file_path, 'r') as fp:
            return yaml.safe_load(fp)

    def validate(self):
        try:
            schema = yamale.make_schema(self.__schema_file_path)
            data = yamale.make_data(self.__settings_file_path)
            yamale.validate(schema, data)
        except YamaleError as e:
            for result in e.results:
                print(f"{result}\n")

    ### Set

    def commit(self):
        settings = self.to_dict()
        self.write(settings)

    def from_dict(self, settings):
        self.__settings = settings
        if settings:
            self.__cli_settings = CLISettings(settings.get('cli'))
            self.__proxy_settings = ProxySettings(settings.get('proxy'))
            self.__remote_settings = RemoteSettings(settings.get('remote'))
            self.__ui_settings = UISettings(settings.get('ui'))

    def load(self):
        self.__load_settings()

    def reset(self):
        self.__detect_paths()
        self.__create_default_file()

    def reset_and_load(self):
        self.reset()
        self.__load_settings()

    def write(self, contents):
        if contents:
            with open(self.__settings_file_path, 'w') as fp:
                yaml.dump(contents, fp, allow_unicode=True)

    ### Helpers

    def __create_default_file(self):
        contents = ''
        with open(SourceDir.instance().settings_template_file_path, 'r') as fp:
            contents = fp.read()

        with open(self.__settings_file_path, 'w') as fp:
            fp.write(contents)

    def __detect_paths(self):
        self.__settings_file_path = os.environ.get(env_vars.AGENT_CONFIG_PATH) or self.__data_dir.settings_file_path
        self.__schema_file_path = SourceDir.instance().schema_file_path

    def __load_settings(self):
        with open(self.__settings_file_path, 'r') as stream:
            try:
                settings = yaml.safe_load(stream)
                if not settings:
                    time.sleep(1) # TODO: Sometimes it takes a bit to read, should look into this
                    settings = yaml.safe_load(stream)

                self.from_dict(settings)
            except yaml.YAMLError as exc:
                Logger.instance().error(exc)

    def __reload_settings(self, event):
        if not self.__load_lock:
            from stoobly_agent.app.proxy.utils.publish_change_service import publish_change

            self.__load_lock = True

            Logger.instance(LOG_ID).debug('Reloading settings')
            self.__load_settings()

            publish_change(statuses.SETTINGS_MODIFIED, self.__settings, sync=True)

        self.__load_lock = False



import pdb
import requests
import time

from stoobly_agent.app.models.adapters.python import PythonRequestAdapterFactory
from stoobly_agent.app.models.schemas.request import Request

class ReplayContext():

  def __init__(self, request: Request):
    self.__end_time = None
    self.__start_time = time.time()

    self.__request: Request = request
    self.__response: requests.Response = None

    self.__sequence = None

  @classmethod
  def from_python_request(cls, request: requests.Request):
    stoobly_request = PythonRequestAdapterFactory(request).stoobly_request()
    return cls(Request(stoobly_request))

  @property
  def end_time(self):
    return self.__end_time

  @end_time.setter
  def end_time(self, v):
    self.__end_time = v 

  # ms
  @property
  def latency(self):
    seconds = self.end_time - self.start_time
    return round(seconds * 1000)

  @property
  def key(self):
    return self.__request.key

  @property
  def request(self):
    return self.__request

  @property
  def response(self):
    return self.__response

  @property
  def response_content(self):
    return self.__response.content

  @property
  def sequence(self):
    return self.__sequence

  @property
  def start_time(self):
    return self.__start_time

  @start_time.setter
  def start_time(self, v):
    self.__start_time = v

  def clone(self):
    return ReplayContext(self.__request.clone())

  def with_response(self, response: requests.Response):
    self.__end_time = time.time()
    self.__response = response
    return self

  def with_sequence(self, sequence):
    self.__sequence = sequence
    return self




import os
import pdb
import threading

from copy import deepcopy
from mitmproxy.http import Request as MitmproxyRequest

from stoobly_agent.app.settings.constants.mode import TEST
from stoobly_agent.app.models.request_model import RequestModel
from stoobly_agent.app.proxy.intercept_settings import InterceptSettings
from stoobly_agent.config.constants.env_vars import ENV
from stoobly_agent.config.constants import lifecycle_hooks, record_policy
from stoobly_agent.lib.logger import Logger

from .constants import custom_response_codes
from .mock.eval_request_service import inject_eval_request
from .record.context import RecordContext
from .record.overwrite_scenario_service import overwrite_scenario
from .record.upload_request_service import inject_upload_request
from .utils.allowed_request_service import get_active_mode_policy
from .utils.response_handler import bad_request, disable_transfer_encoding 
from .utils.rewrite import rewrite_request_response

LOG_ID = 'Record'

###
#
# 1. Rewrites a copy of request and response
# 2. BEFORE_RECORD gets triggered
# 3. AFTER_RECORD gets triggered
#
def handle_response_record(context: RecordContext):
    flow = context.flow
    disable_transfer_encoding(flow.response)

    intercept_settings = context.intercept_settings
    request: MitmproxyRequest = flow.request
    request_model = RequestModel(intercept_settings.settings)

    active_record_policy = get_active_mode_policy(request, intercept_settings)
    Logger.instance(LOG_ID).debug(f"RecordPolicy: {active_record_policy}")

    if active_record_policy == record_policy.ALL:
        __record_request(context, request_model)
    elif active_record_policy == record_policy.FOUND:
        res = inject_eval_request(request_model, intercept_settings)(request, [])

        if res.status_code != custom_response_codes.NOT_FOUND:
            __record_request(context , request_model)
    elif active_record_policy == record_policy.NOT_FOUND:
        res = inject_eval_request(request_model, intercept_settings)(request, [])

        if res.status_code == custom_response_codes.NOT_FOUND:
            __record_request(context, request_model)
    elif active_record_policy == record_policy.OVERWRITE:
        overwrite_scenario(intercept_settings.scenario_key)

        __record_request(context, request_model)
    else:
        if active_record_policy != record_policy.NONE:
            return bad_request(
                flow,
                "Valid env RECORD_POLICY: %s, Got: %s" %
                ([record_policy.ALL, record_policy.FOUND, record_policy.NOT_FOUND], active_record_policy)
            )

def __record_handler(context: RecordContext, request_model: RequestModel):
    flow = context.flow
    flow_copy = deepcopy(flow)
    intercept_settings = context.intercept_settings

    context.flow = flow_copy # Deep copy flow to prevent response modifications from persisting
    rewrite_request_response(flow_copy, intercept_settings.record_rewrite_rules) 
    __record_hook(lifecycle_hooks.BEFORE_RECORD, context)

    inject_upload_request(request_model, intercept_settings)(flow_copy)

    __record_hook(lifecycle_hooks.AFTER_RECORD, context)
    context.flow = flow # Reset flow

def __record_request(context: RecordContext, request_model: RequestModel):
    if os.environ.get(ENV) == TEST:
        __record_handler(context, request_model)
    else:
        thread = threading.Thread(
            target=__record_handler,
            args=[context, request_model]
        )
        thread.start()

def __record_hook(hook: str, context: RecordContext):
    intercept_settings: InterceptSettings = context.intercept_settings
    lifecycle_hooks_module = intercept_settings.lifecycle_hooks

    if hook in lifecycle_hooks_module:
        lifecycle_hooks_module[hook](context)


from typing import List

from .types.proxy_settings import FirewallRule as IFirewallRule

class FirewallRule:

  def __init__(self, firewall_rule: IFirewallRule):
    self.__firewall_rule = firewall_rule

    self.__action = self.__firewall_rule.get('action')
    self.__methods = self.__firewall_rule.get('methods') or []
    self.__modes = self.__firewall_rule.get('modes') or []
    self.__pattern = self.__firewall_rule.get('pattern')


  @property 
  def action(self):
    return self.__action

  @property 
  def methods(self):
    return self.__methods

  @property
  def modes(self):
    return self.__modes

  @property
  def pattern(self):
    return self.__pattern

  def to_dict(self):
    return {
      'action': self.__action,
      'methods': self.__methods,
      'modes': self.__modes,
      'pattern': self.__pattern,
    }


import os
import pdb

class SourceDir:
    _instance = None

    def __init__(self):
        if SourceDir._instance:
            raise RuntimeError('Call instance() instead')
        else:
            self.__src_dir_path = os.path.join(os.path.dirname(os.path.abspath(__file__)), '..')

    @classmethod
    def instance(cls):
        if cls._instance is None:
            cls._instance = cls()

        return cls._instance

    @property
    def path(self):
        return self.__src_dir_path

    @property
    def config_dir_path(self):
        return os.path.join(self.path, 'config')

    @property
    def db_dir_path(self):
        return os.path.join(self.path, 'db')

    @property
    def db_migrations_dir_path(self):
        return os.path.join(self.db_dir_path, 'migrations')

    @property
    def settings_file_path(self):
        return os.path.join(self.config_dir_path, 'settings.yml')

    @property
    def settings_template_file_path(self):
        return os.path.join(self.config_dir_path, 'settings.yml.sample')

    @property
    def schema_file_path(self):
        return os.path.join(self.config_dir_path, 'schema.yml')


import time
import uuid
import pdb

from typing import Union

from stoobly_agent.lib.utils.decode import decode

from .proxy_request import ProxyRequest
from .request_string_control import RequestStringControl

CLRF = "\r\n"

class RequestString:
    ENCODING = 'utf-8'

    __current_time = None

    def __init__(self, proxy_request: ProxyRequest):
        self.__current_time = self.__get_current_time()

        self.lines = []

        if proxy_request:
            self.request = proxy_request.request
            self.proxy_request = proxy_request

            self.__request_line()
            self.__headers()
            self.__body()

        self.request_id = self.__generate_request_id()

    def get(self, **kwargs):
        if kwargs.get('control'):
            return CLRF.join([self.control] + self.lines).encode(self.ENCODING)
        else:
            return CLRF.join(self.lines).encode(self.ENCODING)

    def set(self, s: bytes):
        decoded_s = decode(s, self.ENCODING)
        self.lines = decoded_s.split(CLRF)

    @property
    def control(self):
        control = RequestStringControl()
        control.id = self.request_id
        control.timestamp = self.__current_time
        return control.serialize()

    @control.setter
    def control(self, c: str):
        control = RequestStringControl(c)
        self.request_id = control.id
        self.__current_time = control.timestamp

    def __request_line(self):
        self.lines.append("{} {} HTTP/1.1".format(self.request.method, self.proxy_request.url()))

    def __headers(self):
        headers = self.request.headers

        for name, val in headers.items():
            line = ' '.join([
                "{}:".format(self.__to_header_case(self.__to_str(name))), 
                self.__to_str(val)
            ])
            self.lines.append(line)

    def __body(self):
        self.lines.append("{}{}".format(CLRF, self.request.body))

    def __to_header_case(self, header: str) -> str:
        toks = header.split('_')

        for index, tok in enumerate(toks):
            toks[index] = tok.lower().title()

        return "-".join(toks)

    def __generate_request_id(self):
        return str(uuid.uuid4())

    def __get_current_time(self):
        now = time.time()
        current_time = round(now * (pow(10, 9)))

        return current_time

    def __to_str(self, s: Union[bytes, str]):
        return decode(s)



from mitmproxy.http import Request as MitmproxyRequest, Response as MitmproxyResponse
from stoobly_agent.app.proxy.mitmproxy.request_facade import MitmproxyRequestFacade
from stoobly_agent.app.proxy.mock.hashed_request_decorator import HashedRequestDecorator
from stoobly_agent.app.proxy.mock.request_hasher import RequestHasher

class ORMRequestBuilder():
  def columns_from_mitmproxy_request(self, request: MitmproxyRequest):
    request_facade = MitmproxyRequestFacade(request)
    hashed_request = HashedRequestDecorator(request_facade)

    return {
      'body_params_hash': hashed_request.body_params_hash(),
      'body_text_hash': hashed_request.body_text_hash(),
      'headers_hash': hashed_request.headers_hash(),
      'host': request.host,
      'http_version':self.__http_version(request.http_version),
      'method': request.method,
      'password': request_facade.password,
      'path': request_facade.path,
      'port': request.port,
      'query': request_facade.query_string,
      'query_params_hash': hashed_request.query_params_hash(),
      'scheme': request.scheme,
      'user': request_facade.username,
    }

  def columns_from_mitmproxy_response(self, response: MitmproxyResponse):
    return {
      'response_hash': RequestHasher.instance().hash_text(response.content),
      'response_headers_hash': RequestHasher.instance().hash_params(response.headers)
    }

  def __http_version(self, http_version: str):
    if not isinstance(http_version, str):
      return http_version
    _version = http_version.split('/')[1]
    return float(_version)


import json

from mitmproxy.http import Request as MitmproxyRequest
from urllib.parse import urlparse

from stoobly_agent.lib.logger import Logger
from stoobly_agent.lib.orm.utils.requests_response_builder import RequestsResponseBuilder

LOG_ID = 'ReversePorxy'

def reverse_proxy(request: MitmproxyRequest, service_url: str, options = {}):
    Logger.instance(LOG_ID).debug(service_url)

    uri = urlparse(service_url)

    request.scheme = uri.scheme
    request.host = uri.hostname
    if uri.port:
        request.port = uri.port

def build_response(passed, log):
    body = json.dumps({
        'log': log,
        'passed': passed,
    }).encode()
 
    builder = RequestsResponseBuilder()
    builder.with_header('content-type', 'application/json')
    builder.with_body(body)
    builder.with_status_code(200)

    return builder.build()


from stoobly_agent.app.settings import RemoteSettings
from stoobly_agent.lib.orm.request import Request

from .local_db.query_param_adapter import LocalDBQueryParamAdapter

class QueryParamResourceFactory():

  def __init__(self, settings: RemoteSettings):
    self.__remote_settings = settings

  def local_db(self) -> LocalDBQueryParamAdapter:
    return LocalDBQueryParamAdapter(Request)  


from stoobly_orator.migrations import Migration


class CreateReplayedResponses(Migration):

    def up(self):
        """
        Run the migrations.
        """
        with self.schema.create('replayed_responses') as table:
            table.increments('id')
            table.integer('request_id').unsigned()
            table.foreign('request_id').references('id').on('requests').index()
            table.binary('raw')
            table.integer('status')
            table.integer('latency')
            table.timestamps()

    def down(self):
        """
        Revert the migrations.
        """
        self.schema.drop('replayed_responses')



import requests

from mitmproxy.http import Request as MitmproxyRequest, Response as MitmproxyResponse
from typing import Union

from stoobly_agent.app.models.adapters.python import PythonRequestAdapterFactory, PythonResponseAdapterFactory
from stoobly_agent.app.models.adapters.raw_http_request_adapter import DEFAULT_HTTP_VERSION

class MitmproxyFlowMock():

  def __init__(self):
    self.__request: MitmproxyRequest = None
    self.__response: MitmproxyResponse = None

  @property
  def request(self):
    return self.__request

  @property
  def response(self):
    return self.__response

  @request.setter
  def request(self, _request: Union[MitmproxyRequest, requests.Request]):
    if isinstance(_request, requests.Request):
      self.__request = PythonRequestAdapterFactory(_request).mitmproxy_request(DEFAULT_HTTP_VERSION)
    else:
      self.__request = _request

  @response.setter
  def response(self, _response: Union[MitmproxyResponse, requests.Response]):
    if isinstance(_response, requests.Response):
      self.__response = PythonResponseAdapterFactory(_response).mitmproxy_response()
    else:
      self.__response = _response


import hashlib

from typing import List

from stoobly_agent.lib.cache import Cache
from stoobly_agent.lib.orm.request import Request

SUFFIX = 'last_request_id'

def access_request(session_id: str, request_id: int, timeout = None):
  cache = Cache.instance()
  _last_request_id_key = __last_request_id_key(session_id)
  cache.write(_last_request_id_key, request_id, timeout)

def generate_session_id(query: dict):
  toks = []

  for k in query:
    toks.append(f"{k}#{query[k]}".encode())

  return hashlib.md5(b'.'.join(toks)).hexdigest()

def reset_sessions():
  Cache.instance().clear(f".+\\.{SUFFIX}")

def tiebreak_scenario_request(session_id: str, requests: List[Request]):
  if len(requests) == 0:
    return None

  if len(requests) == 1:
    return requests[0]

  _last_request_id_key = __last_request_id_key(session_id)

  cache = Cache.instance()
  last_request_id_data = cache.read(_last_request_id_key)

  if not last_request_id_data:
    return requests[0]

  last_request_id = last_request_id_data['value']

  for request in requests:
    if request.id > last_request_id:
      return request

  return requests[0]

def __last_request_id_key(_key = None):
  _key = _key or generate_session_id()
  return f"{_key}.{SUFFIX}"


import logging
import logging.config
import os
import pdb
import sys

from typing import Literal

from stoobly_agent.config.constants.env_vars import LOG_LEVEL

class bcolors:
    HEADER = '\033[95m'
    OKBLUE = '\033[94m'
    OKCYAN = '\033[96m'
    OKGREEN = '\033[92m'
    WARNING = '\033[93m'
    FAIL = '\033[91m'
    ENDC = '\033[0m'
    BOLD = '\033[1m'
    UNDERLINE = '\033[4m'

DEBUG = 'debug'
INFO = 'info'
WARNING = 'warning'
ERROR = 'error'
LogLevel = Literal[DEBUG, ERROR, INFO, WARNING]

class Logger:
    _instance = None

    def __init__(self):
        raise RuntimeError('Call instance() instead')

    @classmethod
    def instance(cls, name = None):
        if cls._instance is None:
            cls._instance = cls.__new__(cls)

        cls._instance.load()

        if not name:
            return logging
        else:
            return logging.getLogger(name)

    @classmethod
    def reload(cls):
        if cls._instance is None:
            cls._instance = cls.__new__(cls)

        cls._instance.load()

    def load(self):
        log_level = os.getenv(LOG_LEVEL) or ''

        if log_level.lower() == DEBUG:
            logging.config.dictConfig({'disable_existing_loggers': True, 'version': 1})
            logging.basicConfig(level=logging.DEBUG)
        elif log_level.lower() == WARNING:
            logging.basicConfig(level=logging.WARNING, stream=sys.stderr)
        elif log_level.lower() == ERROR:
            logging.basicConfig(level=logging.ERROR, stream=sys.stderr)
        else:
            logging.basicConfig(level=logging.INFO)


import requests
import urllib

from ..logger import Logger
from .tests_resource import TestsResource

LOG_ID = 'TestResponsesResource'

class TestResponsesResource(TestsResource):
  RESPONSES_ENDPOINT = 'responses'

  def mock(self, test_id: int, **query_params) -> requests.Response:
    url = f"{self.service_url}/{self.TESTS_ENDPOINT}/{test_id}/{self.RESPONSES_ENDPOINT}/mock"

    Logger.instance(LOG_ID).debug(f"{url}?{urllib.parse.urlencode(query_params)}")

    return self.get(url, headers=self.default_headers, params=query_params) 


import os
import pdb

from stoobly_agent.config.constants import env_vars

from .types import FeatureSettings as IFeatureSettings

class FeatureSettings:

  def __init__(self, feature_settings: IFeatureSettings):
    self.__feature_settings = feature_settings or {}
    self.__dev_tools = self.__feature_settings.get('dev_tools')
    self.__exec = self.__feature_settings.get('exec')
    self.__remote = self.__feature_settings.get('remote')

  @property
  def dev_tools(self) -> bool:
    return self.__dev_tools or False

  @property
  def exec(self) -> bool:
    return self.__exec or False

  @property
  def remote(self) -> bool:
    if os.environ.get(env_vars.AGENT_REMOTE_ENABLED):
        return True 
    return self.__remote or False

  @dev_tools.setter
  def api_url(self, v):
    self.__dev_tools = v

  @exec.setter
  def exec(self, v):
    self.__exec = v

  @remote.setter
  def remote(self, v):
    self.__remote = v

  def to_dict(self) -> IFeatureSettings:
    return {
      'dev_tools': self.__dev_tools,
      'exec': self.__exec,
      'remote': self.__remote,
    }


from stoobly_orator.migrations import Migration

from stoobly_agent.lib.orm.response import Response

class AddHttpVersionColumnToResponses(Migration):

    def up(self):
        """
        Run the migrations.
        """
        with self.schema.table('responses') as table:
            table.decimal('http_version').default(1.1)

    def down(self):
        """
        Revert the migrations.
        """
        with self.schema.table('responses') as table:
            table.drop_column('http_version')




import pdb
import requests

from mitmproxy.http import Headers, Response as MitmproxyResponse
from time import time

from stoobly_agent.app.models.adapters.raw_http_request_adapter import DEFAULT_HTTP_VERSION
from stoobly_agent.lib.utils.decode import decode

class MitmproxyResponseAdapter():

  def __init__(self, response: requests.Response):
    self.__timestamp_start = time()

    if hasattr(response, 'raw'):
      self.__http_version = f"HTTP/{response.raw.version / 10.0}"
    else:
      self.__http_version = DEFAULT_HTTP_VERSION

    self.__latency = 0 # Seconds
    self.__response = response
 
  def with_latency(self, latency: int):
    self.__latency = latency
    return self

  @property
  def status_code(self) -> int:
    status_code = self.__response.status_code
    
    if isinstance(status_code, int):
      return status_code

    return int(decode(status_code))

  @property
  def headers(self):
    return Headers(**self.__decode_dict(self.__response.headers))

  @property
  def reason(self):
    return self.__response.reason or ''

  @property
  def timestamp_end(self):
    return self.__timestamp_start + self.__latency

  def adapt(self):
    content = self.__response.content

    res = MitmproxyResponse.make(
      self.status_code,
      content,
      self.headers,
    )

    # After reading .content, .raw.data will be empty
    # However, setting ._body will set .raw.data
    self.__response.raw._body = res.raw_content

    return res

  def __decode_dict(self, d):
    new_d = {}
    for k, v in d.items():
      new_d[decode(k)] = decode(v)
    return new_d


import pdb

from mitmproxy.http import Request
from typing import Union

from stoobly_agent.app.cli.helpers.context import ReplayContext
from stoobly_agent.app.proxy.mitmproxy.request_facade import MitmproxyRequestFacade
from stoobly_agent.app.proxy.mock.context import MockContext
from stoobly_agent.app.proxy.replay.alias_resolver import AliasResolver
from stoobly_agent.app.proxy.replay.body_parser_service import encode_response, is_traversable
from stoobly_agent.app.proxy.replay.rewrite_params_service import build_id_to_alias_map, rewrite_params
from stoobly_agent.app.proxy.test.helpers.endpoint_facade import EndpointFacade
from stoobly_agent.app.proxy.test.helpers.mitmproxy_response_adapter import MitmproxyResponseAdapter
from stoobly_agent.app.proxy.test.helpers.requests_response_adapter import RequestsResponseAdapter
from stoobly_agent.app.proxy.test.helpers.request_component_names_facade import RequestComponentNamesFacade

from stoobly_agent.config.constants import custom_headers
from stoobly_agent.lib.api.endpoints_resource import EndpointsResource
from stoobly_agent.lib.api.interfaces.endpoints import EndpointShowResponse
from stoobly_agent.lib.orm.trace import Trace

from .context_response import TestContextResponse
from .context_abc import TestContextABC

FuzzyContent = Union[dict, list, str]

class TestContext(TestContextABC):
  def __init__(self, replay_context: ReplayContext, mock_context: MockContext):
    self.__flow = mock_context.flow
    self.__intercept_settings = mock_context.intercept_settings
    self.__mock_context = mock_context
    self.__replay_context = replay_context

    mock_response = self.__mock_context.response
    self.__expected_response = RequestsResponseAdapter(mock_response).adapt()

    upstream_response = self.__flow.response
    self.__response = MitmproxyResponseAdapter(upstream_response).adapt()

    self.__log = ''
    self.__passed = None
    self.__skipped = False

    # Optional
    self.__endpoints_resource: EndpointsResource = None

    # Cache
    self.__cached_rewritten_expected_response_content = None
    self.__cached_endpoint_show_response = None
    self.__cached_response_param_names = None

  def with_endpoints_resource(self, resource: EndpointShowResponse):
    self.__endpoints_resource = resource
    return self 

  @property
  def cached_rewritten_expected_response_content(self) -> FuzzyContent:
    if not self.__cached_rewritten_expected_response_content:
      return self.rewritten_expected_response_content

    return self.__cached_rewritten_expected_response_content

  @property
  def decoded_response_content(self) -> FuzzyContent:
    return self.__response.decode_content()

  @property
  def decoded_expected_response_content(self) -> FuzzyContent:
    return self.__expected_response.decode_content()

  @property
  def endpoint(self) -> EndpointFacade:
    resource = self.__endpoints_resource
    endpoint_id = self.mock_request_endpoint_id

    if not resource or not endpoint_id:
      return 

    return EndpointFacade(resource, endpoint_id)

  @property
  def end_time(self):
    return self.__flow.response.timestamp_end

  @property
  def expected_latency(self):
    return self.__mock_context.response.headers.get(custom_headers.RESPONSE_LATENCY)

  @property
  def expected_response(self):
    return self.__expected_response

  @property
  def expected_status_code(self):
    return self.__expected_response.status_code

  @property
  def filter(self):
    return self.intercept_settings.test_filter

  @property
  def flow(self):
    return self.__flow

  @property
  def intercept_settings(self):
    return self.__intercept_settings

  @property
  def lifecycle_hooks(self):
    return self.__intercept_settings.lifecycle_hooks

  @property
  def lifecycle_hooks_path(self):
    return self.__intercept_settings.lifecycle_hooks_path

  @property
  def log(self):
    return self.__log

  @log.setter
  def log(self, v: str):
    self.__log = v

  @property
  def mock_context(self) -> MockContext:
    return self.__mock_context

  @property
  def mock_request_id(self) -> Union[str, None]:
    return self.expected_response.headers.get(custom_headers.MOCK_REQUEST_ID)

  @property
  def mock_request_endpoint_id(self) -> Union[str, None]:
    return self.expected_response.headers.get(custom_headers.MOCK_REQUEST_ENDPOINT_ID)

  @property
  def passed(self):
    return self.__passed

  @passed.setter
  def passed(self, v):
    self.__passed = v

  @property
  def public_directory_path(self):
    return self.__intercept_settings.public_directory_path

  @property
  def replay_context(self) -> ReplayContext:
    return self.__replay_context

  @property
  def request(self) -> Request:
    return self.__flow.request

  @property
  def request_key(self) -> str:
    return self.__replay_context.key

  @property
  def request_headers(self):
    return MitmproxyRequestFacade(self.request).headers

  @property
  def response(self) -> TestContextResponse:
    return self.__response

  @property
  def response_fixtures(self):
    return self.__intercept_settings.response_fixtures

  @property
  def response_fixtures_path(self):
    return self.__intercept_settings.response_fixtures_path

  @property
  def response_param_names(self) -> RequestComponentNamesFacade:
    if self.__cached_response_param_names:
      return self.__cached_response_param_names

    endpoint = self.endpoint

    if not endpoint:
      return RequestComponentNamesFacade([])
  
    self.__cached_response_param_names = endpoint.response_param_names

    return self.__cached_response_param_names

  @property
  def rewritten_expected_response_content(self):
    _decoded_expected_response_content = self.decoded_expected_response_content
    
    trace = self.trace
    if not trace:
      return _decoded_expected_response_content

    if not is_traversable(_decoded_expected_response_content):
      return _decoded_expected_response_content

    if not self.response_param_names:
      return _decoded_expected_response_content

    aliased_response_param_names = self.response_param_names.aliased

    _endpoint = self.endpoint
    if not _endpoint:
      aliases = []
    else:
      aliases = self.endpoint.aliases or []

    if len(aliased_response_param_names) == 0 or len(aliases) == 0:
      return _decoded_expected_response_content

    id_to_alias_map = build_id_to_alias_map(aliases)
    alias_resolver = AliasResolver(trace, self.intercept_settings.alias_resolve_strategy)
    rewrite_params(_decoded_expected_response_content, aliased_response_param_names, id_to_alias_map, alias_resolver)

    self.__cached_rewritten_expected_response_content = _decoded_expected_response_content

    return _decoded_expected_response_content

  @property
  def skipped(self):
    return self.__skipped

  @skipped.setter
  def skipped(self, v: bool):
    self.skipped = v

  @property
  def start_time(self):
    return self.request.timestamp_start

  @property
  def save(self):
    return self.intercept_settings.test_save_results

  @property
  def strategy(self):
    return self.intercept_settings.test_strategy

  @property
  def trace(self) -> Union[Trace, None]:
    headers = self.request.headers
    trace_id = headers.get(custom_headers.TRACE_ID)

    if not trace_id:
      return

    return Trace.find(trace_id)


from stoobly_agent.app.proxy.replay.alias_resolver import AliasResolver
from stoobly_agent.lib.orm.trace_alias import TraceAlias

class AliasContext():

  def __init__(self, alias_resolver: AliasResolver):
    self.__alias_resolver = alias_resolver
    self.__resolved_aliases = []

  @property
  def alias_resolver(self):
    return self.__alias_resolver

  @property
  def resolved_aliases(self):
    return self.__resolved_aliases

  def resolve_alias(self, alias_name: str, value: str):
    trace_alias = self.alias_resolver.resolve_alias(alias_name, value)
    self.resolved_aliases.append(trace_alias)
    return trace_alias

  def assign_alias(self, trace_alias: TraceAlias, value):
    self.__alias_resolver.assign_alias(trace_alias, value)


import time

from .proxy_request import ProxyRequest
from .request_string import RequestString
from .response_string import ResponseString

ENCODING = 'utf-8'
REQUEST_DELIMITTER = '\n🐵🙈🙉\n'.encode(ENCODING)

class JoinedRequest:
    __request_string = None
    __response_string = None

    def __init__(self, proxy_request: ProxyRequest = None):
        self.timestamp = time.time()
        self.proxy_request = proxy_request

        if proxy_request:
            self.__request_string = RequestString(proxy_request)

    @property
    def request_string(self):
        return self.__request_string

    @request_string.setter
    def request_string(self, s: str):
        self.__request_string = s

    @property
    def response_string(self):
        return self.__response_string

    @response_string.setter
    def response_string(self, s: str):
        self.__response_string = s

    def with_response(self, response):
        now = time.time()

        self.__response_string = ResponseString(response, self.__request_string.request_id)

        # milliseconds
        latency = round((now - self.timestamp) * (10 ** 6))
        self.__response_string.with_latency(latency)

        return self

    def build(self):
        if not self.response_string:
            raise Exception('Missing response')

        request_string = self.__request_string.get(control=True)
        response_string = self.__response_string.get(control=True)

        joined = REQUEST_DELIMITTER.join([request_string, response_string])
        return joined



import json
import os
import pdb
import shutil

from typing import Callable

from stoobly_agent.lib.orm.scenario import Scenario

from .snapshot import Snapshot
from .request_snapshot import RequestSnapshot, RequestSnapshotOptions

REQUEST_DELIMITTER = "\n"

class ScenarioSnapshot(Snapshot):
  def __init__(self, uuid: str):
    super().__init__(uuid)

    self.__scenarios_dir_path = self.data_dir.snapshots_scenarios_dir_path
    self.__requests_dir_path = self.data_dir.snapshots_scenario_requests_dir_path
    self.__metadata_backup = None
    self.__requests_backup = None

  @property
  def exists(self):
    return os.path.exists(self.requests_path) and os.path.exists(self.metadata_path)

  @property
  def metadata(self):
    if not os.path.exists(self.metadata_path):
      return {}

    with open(self.metadata_path, 'rb') as fp:
      try:
        return json.loads(fp.read())
      except Exception:
        return {
          'name': self.uuid
        }

  @property
  def metadata_backup(self):
    return self.__metadata_backup

  @property
  def metadata_path(self):
    return os.path.join(self.__scenarios_dir_path, self.uuid)

  @property
  def requests(self):
    requests_file_path = self.requests_path

    if not os.path.exists(requests_file_path):
      return

    raw_requests = []
    handler = lambda request_snapshot: raw_requests.append(request_snapshot.request)
    self.iter_request_snapshots(handler)

    return raw_requests

  @property
  def requests_backup(self):
    return self.__requests_backup

  @property
  def requests_path(self):
    return os.path.join(self.__requests_dir_path, self.uuid)

  def backup_metadata(self):
    if os.path.exists(self.metadata_path):
      with open(self.metadata_path, 'r') as fp:
        self.__metadata_backup = fp.read()

  def backup_requests(self):
    self.__requests_backup = {}

    self.iter_request_snapshots(self.__handle_backup_requests)

  def copy(self, destination):
    self.copy_metadata(destination)
    self.copy_requests(destination)
  
  def copy_metadata(self, dest_dir: str):
    metadata_file_path = self.metadata_path
    return self.copy_file(metadata_file_path, dest_dir)

  def copy_requests(self, dest_dir: str):
    if not os.path.exists(dest_dir):
      os.makedirs(dest_dir, exist_ok=True)

    requests_file_path = self.requests_path

    if os.path.exists(requests_file_path):
      # A request only ever belongs to one scenario
      self.iter_request_snapshots(lambda snapshot: self.__handle_copy_requests(snapshot, dest_dir))

      self.copy_file(requests_file_path, dest_dir)

  def iter_request_snapshots(self, handler: Callable[[RequestSnapshot], None]):
    requests_file_path = self.requests_path

    if os.path.exists(requests_file_path):
      with open(requests_file_path, 'r') as fp:
        self.__requests_backup = {}

        uuids = fp.read().split(REQUEST_DELIMITTER)
        uuids = list(filter(lambda uuid: len(uuid) == 36, uuids))

        for uuid in uuids:
          request_snapshot = RequestSnapshot(uuid)
          handler(request_snapshot)

  def remove(self):
    self.remove_metadata()
    self.remove_requests()

  def remove_metadata(self):
    metadata_file_path = self.metadata_path

    if os.path.exists(metadata_file_path):
      os.remove(metadata_file_path)

  def remove_requests(self):
    requests_file_path = self.requests_path

    if os.path.exists(requests_file_path):
      # A request only every belongs to one scenario
      self.iter_request_snapshots(self.__handle_remove_requests)

      os.remove(requests_file_path)

  def find_resource(self):
    return Scenario.find_by(uuid=self.uuid)

  def write_metadata(self, scenario: Scenario):
    with open(self.metadata_path, 'w') as fp:
      text = json.dumps({
        'description': scenario.description,
        'name': scenario.name,
      })
      fp.write(text)

  def write_requests(self, scenario: Scenario, **options: RequestSnapshotOptions):
    with open(self.requests_path, 'w') as fp:
      uuids = []
      requests = scenario.requests

      for request in requests:
        uuid = request.uuid

        request_snapshot = RequestSnapshot(uuid)
        request_snapshot.write(request, **options)

        uuids.append(uuid)
      
      fp.write(REQUEST_DELIMITTER.join(uuids))

  def __handle_backup_requests(self, request_snapshot: RequestSnapshot):
    self.__requests_backup[request_snapshot.uuid] = request_snapshot.request

  def __handle_copy_requests(self, request_snapshot: RequestSnapshot, dest_dir: str):
    request_snapshot.copy(dest_dir)

  def __handle_remove_requests(self, request_snapshot: RequestSnapshot):
    request_snapshot.remove()


import json
import pdb
import click
import re
import sys

from typing import Union

from stoobly_agent.app.cli.helpers.handle_replay_service import DEFAULT_FORMAT, JSON_FORMAT
from stoobly_agent.app.settings import Settings
from stoobly_agent.lib.api.keys import InvalidOrganizationKey, InvalidProjectKey, InvalidReportKey, InvalidRequestKey, InvalidScenarioKey
from stoobly_agent.lib.api.keys import OrganizationKey, ProjectKey, ReportKey, RequestKey, ScenarioKey

from .trace_aliases import adapt_trace_aliases, Alias, parse_aliases

# Print

def print_invalid_key(resource: str, format = DEFAULT_FORMAT):
  error_message = f"Error: Invalid {resource} key"

  if format == JSON_FORMAT:
    print_error_json_format(error_message)
  else:
    print(error_message, file=sys.stderr) 

def print_error_json_format(error_message):
    output = { 'error': error_message }
    print(json.dumps(output), file=sys.stderr)

# Handle

def handle_invalid_key(resource: str, format = DEFAULT_FORMAT):
  print_invalid_key(resource, format)
  sys.exit(1)

def handle_missing_key(resource: str, format = DEFAULT_FORMAT):
  error_message = f"Error: Missing {resource} key"

  if format == JSON_FORMAT:
    print_error_json_format(error_message)
  else:
    print(error_message, file=sys.stderr) 

  sys.exit(1)

def handle_invalid_alias(expected_alias: Alias, alias_value: str, format = DEFAULT_FORMAT):
  error_message = f"Error: Invalid alias {expected_alias['name']} value, got {alias_value}, expected {expected_alias['value']}"

  if format == JSON_FORMAT:
    print_error_json_format(error_message)
  else:
    print(error_message, file=sys.stderr) 

  sys.exit(1)

def handle_missing_alias(expected_alias: Alias, format = DEFAULT_FORMAT):
  error_message = f"Error: Missing alias {expected_alias['name']}"

  if format == JSON_FORMAT:
    print_error_json_format(error_message)
  else:
    print(error_message, file=sys.stderr)

  sys.exit(1)

# Validate

def validate_organization_key(organization_key) -> OrganizationKey:
  try:
    return OrganizationKey(organization_key)
  except InvalidOrganizationKey:
    handle_invalid_key('organization') if organization_key else handle_missing_key('organization')

def validate_project_key(project_key) -> ProjectKey:
  try:
    return ProjectKey(project_key)
  except InvalidProjectKey:
    handle_invalid_key('project') if project_key else handle_missing_key('project')

def validate_report_key(report_key) -> ReportKey:
  try:
    return ReportKey(report_key)
  except InvalidReportKey:
    handle_invalid_key('report') if report_key else handle_missing_key('report')

def validate_request_key(request_key) -> RequestKey:
  try:
    request_key = RequestKey(request_key)
    project_key = ProjectKey(Settings.instance().proxy.intercept.project_key)

    is_local = ProjectKey.check_is_local(request_key.project_id)
    if project_key.is_local != is_local:
      raise InvalidRequestKey(f"Error: Change to {'local' if is_local else 'remote'} project")

    return request_key
  except InvalidRequestKey:
    handle_invalid_key('request') if request_key else handle_missing_key('request')

def validate_scenario_key(scenario_key) -> ScenarioKey:
  try:
    scenario_key = ScenarioKey(scenario_key)
    project_key = ProjectKey(Settings.instance().proxy.intercept.project_key)

    is_local = ProjectKey.check_is_local(scenario_key.project_id)
    if project_key.is_local != is_local:
      raise InvalidRequestKey(f"Error: Change to {'local' if is_local else 'remote'} project")

    return scenario_key
  except InvalidScenarioKey:
    handle_invalid_key('scenario') if scenario_key else handle_missing_key('scenario')

def validate_aliases(validations, **kwargs) -> Union[Alias, None]:
  assigns = kwargs.get('assign')
  trace_id = kwargs.get('trace_id')

  aliases = parse_aliases(assigns) + adapt_trace_aliases(trace_id)
  aliases_map = {}
  for _alias in aliases:
    aliases_map[_alias.get('name')] = _alias.get('value')

  for validation in validations:
    parsed_validation = parse_aliases([validation])[0]
    name = parsed_validation['name']

    if name not in aliases_map or not aliases_map[name]:
      handle_missing_alias(parsed_validation, kwargs.get('format'))

    if parsed_validation['value'] != None:
      if not re.match(parsed_validation['value'], aliases_map[name]):
        handle_invalid_alias(parsed_validation, aliases_map[name], kwargs.get('format'))

def filter_response(res, status: int):
    if status != 0 and status < 400:
        return False

    print(f"Error: {res}", file=sys.stderr)

    return True

# Prompt

def prompt_project_key(settings: Settings):
  while True:
    project_key = click.prompt('Please enter a project key', type=str)

    try:
      ProjectKey(project_key)
      break
    except InvalidProjectKey:
      print_invalid_key('project')
      print(f"Try `stoobly-agent project list` to find a project key", file=sys.stderr) 

  settings.proxy.intercept.project_key = project_key
  settings.commit()

  return project_key

# Resolve

def resolve_project_key(kwargs: dict, settings: Settings) -> str:
    project_key = kwargs.get('project_key')

    if not project_key:
      project_key = settings.proxy.intercept.project_key

      if not project_key or len(project_key) == 0:
        project_key = prompt_project_key(settings)

    return project_key

def resolve_project_key_and_validate(kwargs: dict, settings: Settings  = Settings.instance()) -> str:
    project_key = resolve_project_key(kwargs, settings)

    validate_project_key(project_key)

    return project_key

def resolve_scenario_key(kwargs: dict, settings: Settings) -> str:
    scenario_key = kwargs.get('key')

    if not scenario_key:
        project_key = validate_project_key(settings.proxy.intercept.project_key)
        data_rule = settings.proxy.data.data_rules(project_key.id)
        scenario_key = data_rule.scenario_key
        #Logger.instance().info(f"No scenario key specified, using {scenario_key}")

    return scenario_key

def resolve_scenario_key_and_validate(kwargs: dict, settings: Settings = Settings.instance()) -> str:
    scenario_key = resolve_scenario_key(kwargs, settings)

    validate_scenario_key(scenario_key)

    return scenario_key


import pdb

from stoobly_agent.lib.orm.request import Request
from stoobly_agent.lib.orm.response import Response
from stoobly_agent.app.proxy.record import REQUEST_DELIMITTER, RequestString, ResponseString

from ..python.response.raw_adapter import RawResponseAdapter
from ..raw_http_response_adapter import RawHttpResponseAdapter
from .response.python_adapter import PythonResponseAdapter

class JoinedRequestStringAdapter():

  def __init__(self, request: Request):
    self.__request = request
    self.__response: Response = request.response

  @property
  def response(self):
    return self.__response

  @response.setter
  def response(self, r: Response):
    self.__response = r

  def adapt(self, base: str = None):
    request_string = RequestString(None) 
    request_string.control = self.__request.control
    request_string.set(self.__request.raw)

    response = self.response
    response_string = ResponseString(None, None)
    response_string.control = response.control
    response_string.set(response.raw)

    toks = [request_string.get(control=True), response_string.get(control=True)]
    if base:
      return REQUEST_DELIMITTER.join([base] + toks)
    else:
      return REQUEST_DELIMITTER.join(toks)

  def decode_response(self):
    response = self.response
    python_response = PythonResponseAdapter(response).adapt()

    adapter = RawHttpResponseAdapter(response.raw)
    adapter.body = python_response.content

    content_encoding_header = 'content-encoding'
    if content_encoding_header in adapter.headers:
      del adapter.headers[content_encoding_header]

    python_response = adapter.to_response()
    response.raw = RawResponseAdapter(python_response).adapt()


import pdb
import pytest

from click.testing import CliRunner

from stoobly_agent.app.models.factories.resource.local_db.request_adapter import (
  LocalDBRequestAdapter,
)
from stoobly_agent.app.models.schemas.request import Request as RequestSchema
from stoobly_agent.app.proxy.replay.trace_context import TraceContext
from stoobly_agent.cli import record
from stoobly_agent.config.constants import alias_resolve_strategy
from stoobly_agent.lib.orm.request import Request
from stoobly_agent.lib.orm.trace_alias import TraceAlias
from stoobly_agent.test.test_helper import DETERMINISTIC_GET_REQUEST_URL, reset

@pytest.fixture(scope='module')
def runner():
    return CliRunner()

class TestTraceContext():

    @pytest.fixture(scope='class', autouse=True)
    def settings(self):
        return reset()

    @pytest.fixture(scope='class')
    def request_show_response(self, runner: CliRunner):
      record_result = runner.invoke(record, [DETERMINISTIC_GET_REQUEST_URL])
      assert record_result.exit_code == 0
      request = Request.last()
      adapter = LocalDBRequestAdapter()
      return adapter.show(request.id, **{
        'body': True,
        'headers': True,
        'query_params': True
      })[0]


    @pytest.fixture(scope='class')
    def endpoint_show_response(self):
      return {
        'aliases': [],
        'body_param_names': [],
        'header_names': [],
        'path_segment_names': [],
        'query_param_names': [],
      }

    class TestWhenFifoStrategy():

      @pytest.fixture(scope='class')
      def trace_context(self):
        context = TraceContext(None, None)
        context.alias_resolve_strategy = alias_resolve_strategy.FIFO
        return context

      @pytest.fixture(scope='class')
      def request_schema(self, request_show_response):
        return RequestSchema(request_show_response)

      class TestWhenHeader():

        @pytest.fixture(scope='class')
        def accept_encodings(self):
          return ['gzip; deflate', 'br']

        @pytest.fixture(scope='class')
        def access_token(self):
          return 'abc123'

        @pytest.fixture(scope='class')
        def aliases(self):
          return [
            {
              'id': 1,
              'name': ':acceptEncoding'
            },
            {
              'id': 2,
              'name': ':accessToken'
            }
          ]

        @pytest.fixture(scope='class')
        def header_names(self, aliases):
          accept_encoding_alias = aliases[0]
          access_token_alias = aliases[1]
          return [
            {
              'alias_id': accept_encoding_alias['id'],
              'id': 1,
              'name': 'Accept-Encoding'
            },
            {
              'alias_id': access_token_alias['id'],
              'id': 2,
              'name': 'Access-Token'
            }
          ]

        @pytest.fixture(scope='class')
        def decorated_endpoint_show_response(self, endpoint_show_response, aliases, header_names):
          endpoint_show_response['aliases'] = aliases
          endpoint_show_response['header_names'] = header_names
          return endpoint_show_response

        @pytest.fixture(scope='class')
        def decorated_request_schema(self, request_schema: RequestSchema, accept_encodings, access_token):
          headers = request_schema.headers
          headers.add('Accept-Encoding', accept_encodings[0])
          headers.add('Accept-Encoding', accept_encodings[1])
          headers.add('Access-Token', access_token)
          request_schema.headers = headers
          return request_schema

        @pytest.fixture(scope='class')
        def accept_encoding_trace_alias(self, trace_context: TraceContext, aliases):
          accept_encoding_alias = aliases[0]
          return trace_context.create_trace_alias(accept_encoding_alias['name'], '1')

        @pytest.fixture(scope='class')
        def access_token_trace_alias(self, trace_context: TraceContext, aliases):
          access_token_alias = aliases[1]
          return trace_context.create_trace_alias(access_token_alias['name'], '1')

        def test_it_resolves_accept_encoding(
          self, 
          trace_context: TraceContext,
          decorated_request_schema: RequestSchema,
          decorated_endpoint_show_response,
          accept_encoding_trace_alias: TraceAlias,
          accept_encodings,
        ):
          trace_context.rewrite_request(decorated_request_schema, decorated_endpoint_show_response)
          trace_alias = TraceAlias.find_by(id=accept_encoding_trace_alias.id)

          assert trace_alias.assigned_to == ','.join(accept_encodings)

        def test_it_resolves_access_token(
          self, 
          trace_context: TraceContext,
          decorated_request_schema: RequestSchema,
          decorated_endpoint_show_response,
          access_token_trace_alias: TraceAlias,
          access_token,
        ):
          trace_context.rewrite_request(decorated_request_schema, decorated_endpoint_show_response)
          trace_alias = TraceAlias.find_by(id=access_token_trace_alias.id)

          assert trace_alias.assigned_to == access_token

      class TestWhenQueryParam():

        @pytest.fixture(scope='class')
        def filter(self):
          return ['1', '2']

        @pytest.fixture(scope='class')
        def page(self):
          return '0'

        @pytest.fixture(scope='class')
        def aliases(self):
          return [
            {
              'id': 1,
              'name': ':filter'
            },
            {
              'id': 2,
              'name': ':page'
            }
          ]

        @pytest.fixture(scope='class')
        def query_param_names(self, aliases):
          filter_alias = aliases[0]
          page_alias = aliases[1]
          return [
            {
              'alias_id': filter_alias['id'],
              'id': 1,
              'name': 'filter'
            },
            {
              'alias_id': page_alias['id'],
              'id': 2,
              'name': 'page'
            }
          ]

        @pytest.fixture(scope='class')
        def decorated_endpoint_show_response(self, endpoint_show_response, aliases, query_param_names):
          endpoint_show_response['aliases'] = aliases
          endpoint_show_response['query_param_names'] = query_param_names
          return endpoint_show_response

        @pytest.fixture(scope='class')
        def decorated_request_schema(self, request_schema: RequestSchema, filter, page):
          query_params = request_schema.query_params

          for v in filter:
            query_params.add('filter', v)
          query_params.add('page', page)
          request_schema.query_params = query_params

          return request_schema

        @pytest.fixture(scope='class')
        def filter_trace_alias(self, trace_context: TraceContext, aliases):
          filter_alias = aliases[0]
          return trace_context.create_trace_alias(filter_alias['name'], '1')

        @pytest.fixture(scope='class')
        def page_trace_alias(self, trace_context: TraceContext, aliases):
          page_alias = aliases[1]
          return trace_context.create_trace_alias(page_alias['name'], '1')

        def test_it_resolves_filter(
          self, 
          trace_context: TraceContext,
          decorated_request_schema: RequestSchema,
          decorated_endpoint_show_response,
          filter_trace_alias: TraceAlias,
          filter
        ):
          trace_context.rewrite_request(decorated_request_schema, decorated_endpoint_show_response)
          trace_alias = TraceAlias.find_by(id=filter_trace_alias.id)

          assert trace_alias.assigned_to == ','.join(filter)

        def test_it_resolves_page(
          self, 
          trace_context: TraceContext,
          decorated_request_schema: RequestSchema,
          decorated_endpoint_show_response,
          page_trace_alias: TraceAlias,
          page,
        ):
          trace_context.rewrite_request(decorated_request_schema, decorated_endpoint_show_response)
          trace_alias = TraceAlias.find_by(id=page_trace_alias.id)

          assert trace_alias.assigned_to == page

      class TestWhenQueryParam():

        @pytest.fixture(scope='class')
        def filter(self):
          return ['1', '2']

        @pytest.fixture(scope='class')
        def page(self):
          return '0'

        @pytest.fixture(scope='class')
        def aliases(self):
          return [
            {
              'id': 1,
              'name': ':filter'
            },
            {
              'id': 2,
              'name': ':page'
            }
          ]

        @pytest.fixture(scope='class')
        def query_param_names(self, aliases):
          filter_alias = aliases[0]
          page_alias = aliases[1]
          return [
            {
              'alias_id': filter_alias['id'],
              'id': 1,
              'name': 'filter'
            },
            {
              'alias_id': page_alias['id'],
              'id': 2,
              'name': 'page'
            }
          ]

        @pytest.fixture(scope='class')
        def decorated_endpoint_show_response(self, endpoint_show_response, aliases, query_param_names):
          endpoint_show_response['aliases'] = aliases
          endpoint_show_response['query_param_names'] = query_param_names
          return endpoint_show_response

        @pytest.fixture(scope='class')
        def decorated_request_schema(self, request_schema: RequestSchema, filter, page):
          query_params = request_schema.query_params

          for v in filter:
            query_params.add('filter', v)
          query_params.add('page', page)
          request_schema.query_params = query_params

          return request_schema

        @pytest.fixture(scope='class')
        def filter_trace_alias(self, trace_context: TraceContext, aliases):
          filter_alias = aliases[0]
          return trace_context.create_trace_alias(filter_alias['name'], '1')

        @pytest.fixture(scope='class')
        def page_trace_alias(self, trace_context: TraceContext, aliases):
          page_alias = aliases[1]
          return trace_context.create_trace_alias(page_alias['name'], '1')

        def test_it_resolves_filter(
          self, 
          trace_context: TraceContext,
          decorated_request_schema: RequestSchema,
          decorated_endpoint_show_response,
          filter_trace_alias: TraceAlias,
          filter
        ):
          trace_context.rewrite_request(decorated_request_schema, decorated_endpoint_show_response)
          trace_alias = TraceAlias.find_by(id=filter_trace_alias.id)

          assert trace_alias.assigned_to == ','.join(filter)

        def test_it_resolves_page(
          self, 
          trace_context: TraceContext,
          decorated_request_schema: RequestSchema,
          decorated_endpoint_show_response,
          page_trace_alias: TraceAlias,
          page,
        ):
          trace_context.rewrite_request(decorated_request_schema, decorated_endpoint_show_response)
          trace_alias = TraceAlias.find_by(id=page_trace_alias.id)

          assert trace_alias.assigned_to == page

      class TestWhenBodyParam():

        @pytest.fixture(scope='class')
        def email(self):
          return 'test@stoobly.com'

        @pytest.fixture(scope='class')
        def aliases(self):
          return [
            {
              'id': 1,
              'name': ':email'
            },
          ]

        @pytest.fixture(scope='class')
        def body_param_names(self, aliases):
          email_alias = aliases[0]
          return [
            {
              'alias_id': email_alias['id'],
              'id': 1,
              'name': 'email',
            }
          ]

        @pytest.fixture(scope='class')
        def decorated_endpoint_show_response(self, endpoint_show_response, aliases, body_param_names):
          endpoint_show_response['aliases'] = aliases
          endpoint_show_response['body_param_names'] = body_param_names
          return endpoint_show_response

        @pytest.fixture(scope='class')
        def decorated_request_schema(self, request_schema: RequestSchema, email):
          headers = request_schema.headers
          headers['content-type'] = 'application/json'
          request_schema.headers = headers
          request_schema.body_params = { 'email': email }
          return request_schema

        @pytest.fixture(scope='class')
        def email_trace_alias(self, trace_context: TraceContext, aliases):
          filter_alias = aliases[0]
          return trace_context.create_trace_alias(filter_alias['name'], '1')

        def test_it_resolves_filter(
          self, 
          trace_context: TraceContext,
          decorated_request_schema: RequestSchema,
          decorated_endpoint_show_response,
          email_trace_alias: TraceAlias,
          email
        ):
          trace_context.rewrite_request(decorated_request_schema, decorated_endpoint_show_response)
          trace_alias = TraceAlias.find_by(id=email_trace_alias.id)

          assert trace_alias.assigned_to == email


from .orm_to_request_transformer import ORMToRequestTransformer
from .orm_to_requests_response_transformer import ORMToRequestsResponseTransformer
from .orm_to_stoobly_request_transformer import ORMToStooblyRequestTransformer
from .orm_to_stoobly_response_transformer import ORMToStooblyResponseTransformer


import os

from typing import List

from stoobly_agent.config.constants import env_vars

from .rewrite_rule import RewriteRule
from .types.proxy_settings import RewriteSettings as IRewriteSettings

class RewriteSettings:
  __rewrite_settings = None

  def __init__(self, rewrite_settings: IRewriteSettings):
    self.__rewrite_settings = rewrite_settings or {}

    self.__rewrite_rules_map = {k: list(map(lambda rule: RewriteRule(rule), v)) for k, v in self.__rewrite_settings.items()}

  def rewrite_rules(self, project_id: str) -> List[RewriteRule]:
    return self.__rewrite_rules_map.get(project_id) or []

  def set_rewrite_rules(self, project_id: str, v: List[RewriteRule]):
    self.__rewrite_rules_map[project_id] = v

  def to_dict(self):
    return {k: list(map(lambda rule: rule.to_dict(), v)) for k, v in self.__rewrite_rules_map.items()}


import requests
import urllib
import pdb

from typing import TypedDict
from ..logger import Logger
from .interfaces import ProjectCreateParams, ProjectsIndexQueryParams
from .stoobly_api import StooblyApi

LOG_ID = 'ProjectsResource'

class ProjectsResource(StooblyApi):
  PROJECTS_ENDPOINT = 'projects'

  def create(self, **params: ProjectCreateParams):
    url = f"{self.service_url}/{self.PROJECTS_ENDPOINT}"

    Logger.instance(LOG_ID).debug(f"{url}?{urllib.parse.urlencode(params)}")

    return self.post(url, headers=self.default_headers, json=params)

  def index(self, **query_params: ProjectsIndexQueryParams) -> requests.Response:
    url = f"{self.service_url}/{self.PROJECTS_ENDPOINT}"

    Logger.instance(LOG_ID).debug(f"{url}?{urllib.parse.urlencode(query_params)}")

    return self.get(url, headers=self.default_headers, params=query_params)

  def show(self, project_id: int, **query_params) -> requests.Response:
    url = f"{self.service_url}/{self.PROJECTS_ENDPOINT}/{project_id}"

    Logger.instance(LOG_ID).debug(f"{url}?{urllib.parse.urlencode(query_params)}")

    return self.get(url, headers=self.default_headers, params=query_params)



import json
import pdb
import uuid

from datetime import datetime
from urllib.parse import urlparse

from stoobly_agent.app.api.simple_http_request_handler import SimpleHTTPRequestHandler
from stoobly_agent.app.cli.helpers.context import ReplayContext
from stoobly_agent.app.models.adapters.orm import JoinedRequestStringAdapter
from stoobly_agent.app.models.helpers.create_request_params_service import build_params
from stoobly_agent.app.models.request_model import RequestModel
from stoobly_agent.app.models.schemas.request import Request
from stoobly_agent.app.proxy.replay.replay_request_service import replay
from stoobly_agent.app.proxy.record import REQUEST_DELIMITTER
from stoobly_agent.app.proxy.record.upload_request_service import upload_staged_request
from stoobly_agent.app.settings import Settings
from stoobly_agent.config.constants import mode
from stoobly_agent.lib.orm.request import Request as OrmRequest
from stoobly_agent.lib.utils.decode import decode

class RequestsController:
    _instance = None

    def __init__(self):
        if self._instance:
            raise RuntimeError('Call instance() instead')
        else:
            self.data = {}

    @classmethod
    def instance(cls):
        if cls._instance is None:
            cls._instance = cls()

        return cls._instance

    def create(self, context: SimpleHTTPRequestHandler):
        body_params = context.params
        if not context.required_params(body_params, ['requests']):
            return

        raw_requests = body_params.get('requests')
        payloads_delimitter = body_params.get('payloads_delimitter') or REQUEST_DELIMITTER
        toks = raw_requests.split(payloads_delimitter)

        if len(toks) % 2 != 0:
            return context.bad_request('Invalid requests format')

        scenario_id = body_params.get('scenario_id')
        scenario_id = int(body_params.get('scenario_id')) if scenario_id else None
        created_requests = []

        for i in range(0, len(toks), 2):
            raw_request = payloads_delimitter.join([toks[i], toks[i + 1]])
            create_params = build_params(raw_request, payloads_delimitter)

            if not create_params:
                # Rollback
                return context.bad_request('Could not parse requests')

            request_model = self.__request_model(context)
            request, status = request_model.create(**{
                **create_params,
                'scenario_id': scenario_id,
                'uuid': str(uuid.uuid4()),
            })

            if context.filter_response(request, status):
                # Rollback
                pass

            created_requests.append(request['list'][0])

        context.render(
            json = {
                'list': created_requests,
                'total': len(created_requests),
            },
            status = 200
        )

    # GET /requests
    def index(self, context: SimpleHTTPRequestHandler):
        request_model = self.__request_model(context)
        requests, status = request_model.index(**context.params)

        if context.filter_response(requests, status):
            return

        context.render(
            json = requests,
            status = 200
        )

    # GET /requests/:id
    def get(self, context: SimpleHTTPRequestHandler):
        context.parse_path_params({
            'id': 1
        })

        request_model = self.__request_model(context)
        request, status = request_model.show(context.params.get('id'))

        if context.filter_response(request, status):
            return

        context.render(
            json = request,
            status = 200
        )

    # POST /requests/:id/push
    def push(self, context: SimpleHTTPRequestHandler):
        context.parse_path_params({
            'id': 1,
        })

        body_params = context.params

        if not context.required_params(body_params, ['project_key']):
            return

        request = OrmRequest.find_by(id=body_params.get('id'))

        if not request:
            return context.not_found(f"Could not find request {body_params.get('id')}")

        request_model = RequestModel(Settings.instance())
        request_model.as_remote()
        res = upload_staged_request(
            request,
            request_model,
            body_params.get('project_key'),
            body_params.get('scenario_key')
        )

        if not res:
            return context.internal_error()

        request.update(pushed_at = datetime.now())
        request_model.as_local()
        request, status = request_model.show(request.id)

        if context.filter_response(request, status):
            return

        # This can error out if res is not properly formatted
        # The request has a local ID (the ID of the request that was pushed)
        # and a remote ID (the ID of the request that was created)
        try:
            if len(res['list']):
                request['remote_id'] = res['list'][0]['id']
        except Exception:
            pass

        context.render(
            json = request,
            status = 200
        )

    # PUT /requests/:id
    def update(self, context: SimpleHTTPRequestHandler):
        context.parse_path_params({
            'id': 1
        })

        request_params = context.params.get('request')
        if request_params.get('scenario_id') == -1:
            request_params['scenario_id'] = None

        request_id = context.params.get('id')
        request_model = self.__request_model(context)
        request, status = request_model.update(request_id, **request_params)

        if context.filter_response(request, status):
            return

        context.render(
            json = request,
            status = 200
        )

    # DELETE /requests/:id
    def destroy(self, context: SimpleHTTPRequestHandler):
        context.parse_path_params({
            'id': 1
        })

        request_id = context.params.get('id')

        request_model = self.__request_model(context)
        request, status = request_model.destroy(request_id)

        if context.filter_response(request, status):
            return

        context.render(
            plain = '',
            status = 200
        )

    # GET /requests/:id/replay
    def replay(self, context: SimpleHTTPRequestHandler):
        context.parse_path_params({
            'id': 1
        })
        project_id = int(context.params.get('project_id'))
        request_id = int(context.params.get('id'))

        request_model = self.__request_model(context)
        request_response, status = request_model.show(request_id, **{
            'body': True,
            'headers': True,
            'project_id': project_id,
            'query_params': True,
        })

        if context.filter_response(request_response, status):
            return

        replay_context = ReplayContext(Request(request_response))
        self.__replay(context, replay_context)

    # PUT /requests/send
    def send(self, context: SimpleHTTPRequestHandler):
        body_params = context.params

        headers = []
        try:
            headers = json.load(decode(body_params.get('headers')))
        except Exception as e:
            pass

        url = urlparse(decode(body_params.get('url')))
        request_response = {
            'body': decode(body_params.get('body')),
            'headers': headers,
            'method': decode(body_params.get('method')),
            'path': url.path,
            'password': url.password,
            'port': url.port,
            'query': url.query,
            'url': url.geturl(),
            'username': url.username,
        }

        replay_context = ReplayContext(Request(request_response))
        self.__send(context, replay_context)

    def download(self, context: SimpleHTTPRequestHandler):
        context.parse_path_params({
            'id': 1
        })
        request_id = int(context.params.get('id'))
        format = context.params.get('format')
        request = OrmRequest.find(request_id)

        if not request:
            return context.not_found()

        if format == 'gor':
            filename = f"REQUEST-{int(datetime.now().timestamp())}.gor"
            text = JoinedRequestStringAdapter(request).adapt()

            context.render(
                download = text,
                filename = filename,
                headers = {
                    'Access-Control-Expose-Headers': 'Content-Disposition'
                },
                status = 200
            )
        else:
            return context.bad_request('Invalid format')

    def __request_model(self, context: SimpleHTTPRequestHandler):
        access_token = context.headers.get('access-token')
        request_model = RequestModel(Settings.instance(), access_token=access_token)
        return request_model

    def __replay(self, context: SimpleHTTPRequestHandler, replay_context: ReplayContext):
        options = {}
        if bool(context.params.get('save')):
            options['save'] = True

        self.__send(context, replay_context, **options)

    def __send(self, context: SimpleHTTPRequestHandler, replay_context: ReplayContext, **replay_options):
        res = replay(replay_context, { **replay_options, 'mode': mode.REPLAY })

        if 'Access-Control-Allow-Origin' in res.headers:
            del res.headers['Access-Control-Allow-Origin']

        context.render(
            data = res.raw.data if hasattr(res, 'raw') else res.content,
            headers = res.headers,
            status = res.status_code
        )



import os
import pdb
import sys

from dataclasses import dataclass
from typing import Union

SCAFFOLD_HOSTS_DELIMITTER_BEGIN = "##### STOOBLY SCAFFOLD HOSTS BEGIN #####\n"
SCAFFOLD_HOSTS_DELIMITTER_END = "##### STOOBLY SCAFFOLD HOSTS END #####\n"

class HostsFileManager():

  @dataclass
  class IpAddressToHostnames:
    ip_address: str
    hostnames: list[str]

  # Split IP address and hostnames. Don't include inline comments
  def __split_hosts_line(self, line: str) -> list[str]:
    ip_addr_hosts_split = line.split('#')[0].split()
    return ip_addr_hosts_split

  def get_hosts_file_path(self) -> str:
    file_path = '/etc/hosts'
    if not os.path.exists(file_path):
      print(f"Error: File {file_path} not found.", file=sys.stderr)
      sys.exit(1)
    return file_path

  # Parses hosts file and returns a mapping of IP address to hostnames in a list.
  def get_hosts(self) -> list[IpAddressToHostnames]:
    hosts_file_path = self.get_hosts_file_path()

    if not hosts_file_path:
      return []

    with open(hosts_file_path, 'r') as f:
      hostlines = f.readlines()

    # Skip comments and empty lines
    hostlines = [line.strip() for line in hostlines
                 if not line.startswith('#') and line.strip() != '']

    hosts = []
    for line in hostlines:
      ip_addr_hosts_split = self.__split_hosts_line(line)
      ip_address = ip_addr_hosts_split[0]
      hostnames = ip_addr_hosts_split[1:]
      ipAddressToHostnames = self.IpAddressToHostnames(ip_address, hostnames)

      hosts.append(ipAddressToHostnames)

    return hosts

  def find_host(self, hostname) -> Union[IpAddressToHostnames, None]:
    hosts = self.get_hosts()

    for mapping in hosts:
      if ((mapping.ip_address == '0.0.0.0' or mapping.ip_address == '127.0.0.1') and
        hostname in mapping.hostnames):

        return mapping

    return None

  def install_hostnames(self, hostnames: list[str]) -> None:
    hosts_file_path = self.get_hosts_file_path()

    self.__add_lines_between_markers(
      hosts_file_path, SCAFFOLD_HOSTS_DELIMITTER_BEGIN, SCAFFOLD_HOSTS_DELIMITTER_END, hostnames
    )

  def uninstall_hostnames(self, hostnames: list[str] = []) -> None:
    hosts_file_path = self.get_hosts_file_path()

    self.__remove_lines_between_markers(
      hosts_file_path, SCAFFOLD_HOSTS_DELIMITTER_BEGIN, SCAFFOLD_HOSTS_DELIMITTER_END, hostnames
    ) 

  def __remove_lines_between_markers(self, file_path, start_marker, end_marker, hostnames = []):
    with open(file_path, "r") as file:
      lines = file.readlines()

    filtered_lines = []
    index = 0

    # Continue until we reach start_marker
    for line in lines:
      index += 1

      if start_marker in line:
        break

      filtered_lines.append(line)

    # Continue until we reach end_marker
    found_hostnames = {}
    section = []
    for line in lines[index:]:
      index += 1

      if end_marker in line:
        break
        
      found = False
      for hostname in hostnames:
        if hostname in line:
          if hostname not in found_hostnames:
            print(f"Removing hostname {hostname}")

          found_hostnames[hostname] = True
          found = True

      if not found: 
        section.append(line)

    # If there are still lines in the section
    if len(section):
      filtered_lines.append(start_marker)
      section.append(end_marker)    
      filtered_lines += section

    for line in lines[index:]:
      filtered_lines.append(line)

    with open(file_path, "w") as file:
      file.writelines(filtered_lines)

  def __add_lines_between_markers(self, file_path, start_marker, end_marker, hostnames = []):
    with open(file_path, "r") as file:
      lines = file.readlines()

    filtered_lines = []
    index = 0

    # Continue until we reach start_marker
    for line in lines:
      index += 1
      if start_marker in line:
        break

      filtered_lines.append(line)

    # If no empty line before start_marker, add one
    if len(filtered_lines):
      last_line = filtered_lines[-1]

      if last_line != "\n":
        filtered_lines.append("\n")

    filtered_lines.append(start_marker)

    # Continue until we reach end_marker
    found_hostnames = {}
    for line in lines[index:]:
      index += 1

      if end_marker in line:
        break
        
      found = False
      for hostname in hostnames:
        if hostname in line:
          filtered_lines.append(line)
          found_hostnames[hostname] = True
          found = True

      if not found:
        filtered_lines.append(line)
      
    for hostname in hostnames:
      if hostname in found_hostnames:
        continue

      print(f"Installing hostname {hostname}")
      filtered_lines.append(f"127.0.0.1 {hostname}\n")
      filtered_lines.append(f"::1       {hostname}\n")

    filtered_lines.append(end_marker)    

    for line in lines[index:]:
      filtered_lines.append(line)

    with open(file_path, "w") as file:
      file.writelines(filtered_lines)
